
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"
#include "QuadTree.h"
#include "svpmap.h"

#include "ImagePage.h"
#include "misc/bb.h"


static char *
makeKey(VoidPtr value, char buffer[24])
{
	sprintf(buffer, "%d", value);
	return buffer;
}

QuadLeaf::QuadLeaf()
{
	dllistCtor(&this->list);
}

QuadLeaf::~QuadLeaf()
{
	dllistDtor(&this->list);
}

void
QuadLeaf::addItem(VoidPtr item)
{
	dllistAppend(&this->list, item);
}

void
QuadLeaf::deleteItem(VoidPtr item)
{
	Ref r;
	if(r = dllistSeek(&this->list, item))
		dllistRemove(&this->list, r);
}

void
QuadLeaf::addMyGuys(stringVoidPtrMap *return_map)
{
	Ref r;
	VoidPtr value;
	
	char buffer[24];
	char *key;
	
	for(r = dllistFirst(&this->list); r; dllistNext(&this->list, &r)){
		value = dllistContents(&this->list, r);
		key = makeKey(value, buffer);
		if(svmapContainsP(return_map, key))
			continue;
		svmapSet(return_map, key, value);
	}
}

QuadNode::QuadNode(const iBox& dims, int depth)
{
	int i;
	int j;
	
	this->bounds = dims;
	if(depth == 0) {
		this->quarters = 0;
		this->leaf = new QuadLeaf();
	} else {
		this->leaf = 0;
		this->quarters = new QuadNodePtr[4];
		iBox new_dims;
		new_dims.w = dims.w >> 1;
		new_dims.h = dims.h >> 1;
		for(i = 0; i < 2; i++)
			for(j = 0; j < 2; j++) {
				int index;
				new_dims.x = dims.x + i * (dims.w >> 1);
				new_dims.y = dims.y + j * (dims.h >> 1);
				index = i * 2 + j;
				this->quarters[index] = new QuadNode(new_dims, depth - 1);
			}
	}
}

QuadNode::~QuadNode()
{
	int i;
	
	if(this->quarters) { 
		for(i = 0; i < 4; i++) {
			delete this->quarters[i];
		}
		delete this->quarters;
	}
	if(this->leaf) {
		delete this->leaf;
	}
}


void
QuadNode::addItem(VoidPtr item, const iBox& rect)
{
	int i;
	
	if(overlapBoxes(this->bounds, rect))
		if(this->leaf) 
			this->leaf->addItem(item);
		else if (this->quarters) 
			for(i = 0; i < 4; i++)
				this->quarters[i]->addItem(item, rect);
}

void
QuadNode::deleteItem(VoidPtr item, const iBox& rect)
{
	int i;
	
#if 0
	if(overlapBoxes(this->bounds, rect))
		if(this->leaf) 
			this->leaf->deleteItem(item);
		else if (this->quarters) 
			for(i = 0; i < 4; i++)
				this->quarters[i]->deleteItem(item, rect);
#else
	// bogus -- quick fix for an error
	// either a bug in updating the box
	// or in removing from the quadtree
	// could be detected by writing a search routine
	if(this->leaf) 
		this->leaf->deleteItem(item);
	else if (this->quarters) 
		for(i = 0; i < 4; i++)
			this->quarters[i]->deleteItem(item, rect);
#endif
}

void
QuadNode::findOverlappingItems(const iBox& b, stringVoidPtrMap *return_map)
{
	int i;
	
	if(overlapBoxes(this->bounds, b))
		if(this->leaf) 
			this->leaf->addMyGuys(return_map);
		else if (this->quarters) 
			for(i = 0; i < 4; i++)
				this->quarters[i]->findOverlappingItems(b, return_map);
}

QuadTree::QuadTree(BoundsFunction input_function,
				   void *input_user_data)
	: node(0),
	  function(input_function),
	  user_data(input_user_data)
{
}

QuadTree::~QuadTree()
{
	if(this->node)
		delete this->node;
}

void
QuadTree::createTree(iBox dimensions, int depth)
{
	if(this->node)
		delete this->node;
	
	this->node = new QuadNode(dimensions, depth);
}

void
QuadTree::addItems(DLList *items)
{
	VoidPtr entity;
	iBox box;
	Ref r;
	
	if(this->node == 0)
		return;
	
	for(r = dllistFirst(items); r; dllistNext(items, &r)) {
		entity = (VoidPtr) dllistContents(items, r);
		box = (*this->function) (this->user_data, entity);
		this->node->addItem(entity, box);
	}
}

void
QuadTree::addItem(VoidPtr entity)
{
	iBox box;
	
	if(this->node == 0)
		return;
	
	box = (*this->function) (this->user_data, entity);
	this->node->addItem(entity, box);
}

void
QuadTree::deleteItems(DLList *items)
{
	VoidPtr entity;
	iBox box;
	Ref r;
	
	if(this->node == 0)
		return;
	
	for(r = dllistFirst(items); r; dllistNext(items, &r)) {
		entity = (VoidPtr) dllistContents(items, r);
		box = (*this->function)(this->user_data, entity);
		this->node->deleteItem(entity, box);
	}
}

void
QuadTree::deleteItem(VoidPtr entity)
{
	iBox box;
	
	if(this->node == 0)
		return;
	
	box = (*this->function)(this->user_data, entity);
	this->node->deleteItem(entity, box);
}

void
QuadTree::findOverlappingItems(const iBox& box, DLList *return_list)
{
	if(this->node == 0)
		return;
	
	stringVoidPtrMap map;
	Ref r;
	VoidPtr value;
	
	svmapInit(&map, 0, 10001);
	
	this->node->findOverlappingItems(box, &map);
	
	// copy map into the list
	for(r = svmapFirst(&map); r; r = svmapNext(&map, r)){
		value = svmapContents(&map, r);
		dllistAppend(return_list, value);
	}
	
	svmapCleanup(&map);
}

void
bbPrint(const iBox& bb)
{
	printf("(%3d, %3d) %3dx%3d\n", bb.x, bb.y, bb.w, bb.h);
}

void
QuadTree::print()
{
	if(this->node == 0)
		printf("Empty tree");
	else 
		this->node->print(0);
}

void 
QuadLeaf::print(int indent_level)
{
	int j;
	Ref r;
	DLList *l = &this->list;
	VoidPtr value;
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)){
		value = dllistContents(l, r);
		for(j = 0; j <= indent_level; j++)
			printf("\t");
		printf("%x\n", value);
	}
}

void
QuadNode::print(int indent_level)
{
	int i, j;
	
	for(j = 0; j < indent_level; j++)
		printf("\t");
	bbPrint(this->bounds);
	
	if(this->leaf){
		this->leaf->print(indent_level);
	} else {
		
		for(i = 0; i < 4; i++)
			this->quarters[i]->print(indent_level+1);
	}
}


void 
QuadNode::debugFind(VoidPtr entity, const iBox&  b)
{
	int i;
	
	if(bounds.w == 0 || overlapBoxes(this->bounds,b)){
		if(this->leaf){
			if(dllistSeek(&this->leaf->list, entity))
				printf("Found entity as expected in node %3d, %3d %3dx%3d\n",
					   this->bounds.x, this->bounds.y, 
					   this->bounds.w, this->bounds.h);
			else 
				printf("Entity not found as expected in node %3d, %3d %3dx%3d\n",
					   this->bounds.x, this->bounds.y, 
					   this->bounds.w, this->bounds.h);
		} else {
			if (this->quarters) 
				
				for(i = 0; i < 4; i++)
					this->quarters[i]->debugFind(entity, b);
		}
	} 
}
void
QuadTree::debugFind(VoidPtr entity, int obey_bounds)
{
	iBox bounds;
	
	if(this->node == 0)
		return;
	
	if(obey_bounds){
		bounds = (*this->function) (this->user_data, entity);
		printf("Entity bounds %3d, %3d %3dx%3d\n", 
			   bounds.x, bounds.y, bounds.w, bounds.h);
	} else {
		bounds.w = 0;
	}
	
	this->node->debugFind(entity, bounds);
}
