
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"
#include "dafs.h"
#include "smemory.h"

#include "MainWindow.h"
#include "Mode.h"
#include "RTEMode.h"
#include "ViewCanvas.h"
#include "EntityAux.h"
#include "RTEPage.h"
#include "RTElement.h"
#include "misc/Unicode.h"
#include "scroll.h"
#include "keyboard/KeyMap.h"
#include "keyboard/KeyBoard.h"
#include "SelectionManager.h"
#include "misc/XArg.h"
#include "rte/Rectangle.h"
#include "dafs.h"
#include "RTEUniEditor.h"
#include "IllCanvas.h"
#include "Application.h"

#define MIN_ELEMENT_WIDTH 10
#define DEFAULT_X_MARGIN 4 
#define DEFAULT_Y_MARGIN 0

RTElement::RTElement(RTEPage *input_page,iEntityPtr input_entity,Coord x, Coord y, Coord w, Coord h, 
                     float *x_expansion_factor, float *y_expansion_factor)
	: page(input_page),
	  ChildRectangle(x,y,w,h),
	  EntityFace(input_page,input_entity),
	  canvas(0)
{
	MainWindow *mw = input_page->mode->main_window;
	RTEPage *rte_page = (RTEPage *) input_page;
	int calc_w, calc_h;

	this->canvas = new IllCanvas(0, mw->app->app_context,
						   rte_page->bb, 0, 0,
						   mw);
	this->display_widget = new RTEUniEditor(this->canvas, FALSE, mw, this);
	this->canvas->setClient(this->display_widget);
	this->parent = NULL;
	this->setPositionNotValid();
	this->display_widget->setXMargin(DEFAULT_X_MARGIN);
	this->display_widget->setYMargin(DEFAULT_Y_MARGIN);
	this->display_widget->setEntity(input_entity);

	iLBox calc_box;
	calc_box = this->display_widget->getDesiredSize();
	calc_w = calc_box.w;
	calc_h = calc_box.h;
#if 0
	this->display_widget->calculateSize(&calc_w, &calc_h);
#endif

	*x_expansion_factor = MAX(1.0, (float)calc_w/(float)w);
	*y_expansion_factor = MAX(1.0, (float)calc_h/(float)h);
	this->setPosition(x, y);
	this->setSize(calc_w, calc_h);
	const int redraw = TRUE, allow_overlap = TRUE;
	if(!this->becomeChildOf((ParentRectangle *) input_page, redraw, allow_overlap))
		halt("RTElement failed to become child of page\n");
	this->setPosition(this->getX(), this->getY());
	XtManageChild(this->display_widget->getWidget());
}

RTElement::~RTElement()
{
	delete this->display_widget;
	delete this->canvas;
}

void
RTElement::setSize(Dimension w, Dimension h)
{
	if(!this->changeSizeTo(w, h))
		halt("RTElement::setSize failed\n");
}

void
RTElement::setPosition(Position x, Position y)
{
	if(!this->moveTo(x, y))
		halt("RTElement::setPosition failed\n");
}

int
RTElement::changeSizeTo(Coord w, Coord h, int redraw)
{
	w = MAX(MIN_ELEMENT_WIDTH, w);
	// this order should be right because the parent widget
	// will change size in time to allow the child to legally change
	// size
	const int result = this->ChildRectangle::changeSizeTo(w, h, redraw);
	if(result) {
		// gnarley - doesn't the UniEditor::setSize do this?
		XArg args;
		args.reset();
		args.setArg(XmNwidth, w);
		args.setArg(XmNheight, h);
		args.setValues(this->display_widget->getWidget());
		this->display_widget->UniEditor::setSize(w, h);
	}
	return result;
}

int 
RTElement::moveTo(Coord x, Coord y, int squeeze, int allow_overlap)
{
	const int result = this->ChildRectangle::moveTo(x, y, squeeze, allow_overlap);
	if(result) {
		XArg args;
		args.reset();
		args.setArg(XmNx, x);
		args.setArg(XmNy, y);
		args.setValues(this->display_widget->getWidget());
	}
	return result;
}

void
RTElement::showCursor()
{
	iEntityPtr from_entity, to_entity;
	int from_index, to_index;
	int success;
	RTEPage *page = (RTEPage*)this->page;
	int from_line, from_col;
	int to_line, to_col;

	page->getSM()->convertToRange();

	page->getSM()->getRange(from_entity, from_index,
						  to_entity, to_index);

	RTElement *from_element = 0;
	RTElement *to_element = 0;

	// first -- decide to ignore straddling selections
	if(from_entity) {
		assert(to_entity);
		from_element = page->findElement(from_entity);
		to_element = page->findElement(to_entity);
	}

	if(from_element && (from_element == to_element)  && (from_element == this)) {
		RTEUniEditor *from_editor = (RTEUniEditor *) from_element->display_widget;
		success = from_editor->convertDAFS2Line(from_entity, from_index,
									  &from_line, &from_col);

		success = success && from_editor->convertDAFS2Line(to_entity, to_index,
														   &to_line, &to_col);
		if(success) {
			// override our own setSelection, which would try
			// send an unnecssary message to the SM
			if((from_line != 0 || to_line != 0  || from_col != 0 || to_col != 0) &&
				ucslen(from_editor->getString()) > 0){
				if(from_line == to_line && from_col == to_col)
					from_editor->UniEditor::setCursor(from_line, from_col);
				else
					from_editor->UniEditor::setSelection(from_line, from_col,
											   to_line, to_col);
			}
			// barebones setSelection isn't doing this and no
			// fear of previous active_editor
			page->setActiveEditor(from_editor);
		} else {
			page->getSM()->clear();
			this->display_widget->setCursor(0,0);
		}
	} else {
		this->display_widget->setCursor(0,0);
	}
}

void
RTElement::move()
{
}
void RTElement::willMove()
{
}
void RTElement::recognize()
{
}
void
RTElement::destroy()
{
}
void
RTElement::update()
{
}
void
RTElement::select()
{
}
void
RTElement::unSelect()
{
}
MainWindow*
RTElement::getMainWindow()
{
	RTEPage *page = (RTEPage*)this->page;
	return page->mode->main_window;
}
int RTElement::drawElement()
{
	// display_widget->draw();
	return TRUE;
}

int RTElement::drawRegion(Rectangle *r)
{
	return TRUE;
}

void 
RTElement::deleteCheap()
{
	delete this;
}

void 
RTElement::deleteAndUpdate()
{
}

void 
RTElement::create()
{
}

void 
RTElement::willUpdate()
{
}
int RTElement::drawBorder()
{
	return TRUE;
}


int RTElement::pressKey(iUCode key_value)
{
	return TRUE;
}

int RTElement::pressButton(XEvent *event)
{
	return TRUE;
}

int RTElement::releaseButton(XEvent *event)
{
	return TRUE;
}
