/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#ifndef _H_dafstype_
#define _H_dafstype_

#include "rafport.h" /* for the definition of int4 */

/*
// contains definitions of the data types used by external functions
// other data types are hidden in entity.h
// also has all the tags for reading and writing -- they probably
// belong elsewhere
*/

/* not wchar or any such because
   that could be 2 or 4 bytes -- how annoying 
*/
typedef unsigned short iUCode;

typedef unsigned char iToken;

#if 1
typedef int4 iFixedPoint;
/* fixed point conversions etc. */
#define i_INT2_TO_FP(x) (((iFixedPoint)(x))<<16)
#define i_FP_TO_INT2(x) ((int2)((x)>>16))
#define i_FP_FRAC(x) ((x)&0xFFFFL)
#define iFP_MAX 0x7fffffff
#define iFP_MIN (-iFP_MAX-1)

#define iFP_HALF (1L<<15)
#define iFP_ONE (1L<<16)

#define i_D_TO_FP(x) ((iFixedPoint)((x)*65536.0))
#define i_FP_TO_D(x) ((double)((x)/65536.0))

/* this macro should probably be a function that does the multiply better */
#define i_MUL_FP(x,y) ( ((x)/256)*((y)/256) )

#else

/* leave this in for tests on using double instead of fixed point */

#include <math.h>
#define iFP_MAX 0x7fffffff
#define iFP_MIN (-iFP_MAX-1)

typedef double iFixedPoint;
#define i_INT2_TO_FP(x) ((iFixedPoint)(x))
#define i_FP_TO_INT2(x) ((int2)(x))
#define i_FP_FRAC(x) (((x)-floor(x)))
#define iFP_ONE 1.0
#define iFP_HALF 0.5

#define i_D_TO_FP(x) (x)
#define i_FP_TO_D(X) (x)
#define i_MUL_FP(x,y) ((x)*(y))

#endif

enum iDTag { iDNo, iDPosSet, iDString, iD1PosSet, iDUserData,
			 iDConf, iDShort, iDUString };

/*  iDTag is the entity's content type, returned by
    i_GetEDataType().
	
    iDNo says there is no content or the content is unset.
	
    iDPosSet says the content is a possibility set which can
    contain any number of character/confidence pairs.  An
    example is given in one of the documents below.
	
    iDString says the content is a data string.  An entity
    of type "word" might have a content of "Word".  An
    example is given below.
	
    iD1PosSet is much like a poset with a single entry.
    Internally, it uses the four bytes which would, in a
    DPosSet, be a pointer to a structure containing the
    possibility set, and puts the character in the first
    two bytes and the confidence in the second two bytes.
    D1PosSet is used in i_SetGlyph() and i_GetGlyph();
    The various ways of handling confidence values, of
    dealing with glyphs which may be labelled with ASCII
    (or Unicode) strings, and the like are not yet fully
    resolved.
	*/

enum iATag { iABox, iAOrthogonList, iABoxList };

enum iPTag { iPAny,iPNo,iPBool,iPLong,iPString,iPUString,
			 iPUserData, iPUserIData, iPUserPtr };


/*  if iLogicalType is iAnd, the children of this entity are
    to be read sequentially.
	
    if iLogicalType is iOr, the children of this entity are to
    be read as alternatives.
	*/
enum iLogicalType {
	iAnd, iOr
};


/*  This determines what kind of DAFS document will be written out. */
enum iDAFSWriteMode {
	DAFSB,
	DAFSU,
	DAFSA
};

/* The currently allowed image compression data formats. */
enum iDAFSCompType {
	NoComp,
	CCITT3,
	CCITT4,
	CCITT32D /* supported for read but not write */
};

enum iSearchType {
	ALL,
	POSSET,
	PROPERTY,
	NOCHILD,
	NOTOR,
	DATAEN
};


/* Errors returned by library routines when something goes wrong. */
enum iDAFSError {
	DAFSOK               = 0,
	FirstError           =-99,
	MallocFailError      ,
	NonExistentFileError ,
	WriteLockedFileError ,
	ProtectedFileError   ,
	DiskFullError        ,
	DAFSWriteModeError   ,
	WrongPropType        ,
	WriteError			 ,
	ReadError			 ,
	BadFileType			 ,
	UserStop			 ,
	NonExistentProp		 ,
	UnknownDAFSError	 ,
	NotEnoughData		 ,
	NotInitialized       ,
	BadRotError          ,
	CantRotateError      ,
	BadScaleError        ,
	WrongImageFormat     ,
	WrongFileType        ,
	AlreadyOrEntity      ,
	DiffImageType        ,
	BadBoundsError       ,
	BadEntityPassed		 ,
	CantLimitError		 ,
	OutOfBoundsError	 ,
	PosSetShareedError  ,
	FailedTest           ,
	NoMoreData			,
	CantWrite32D	,
	CantParseString	,
	LastError            ,	 /* make sure that this is at the end of errors */
};


/* just to save space -- instead of an int2 or and int4 */
typedef unsigned char iTypeIndex;



/* messages are sent to the callback when certain operations happen
   the idea is that the programmer will be making a user interface 
   update
*/
enum iMess {
	mUpdate,
	mDestroy,
	mCreate,
	mWillUpdate,
	mWillMove,
	mMove,
	mDestImg
};


/* so that the storage size is consistent across platforms */
#ifdef SYS_OS_WIN16
typedef long iDAFSError;
typedef long iDTag;
typedef long iATag;
typedef long iPTag;
typedef long iLogicalType;
typedef long iDAFSWriteMode;
typedef long iDAFSCompType;
typedef long iSearchType;
typedef long iMess;
#else
typedef enum iDAFSError iDAFSError;
typedef enum iDTag iDTag;
typedef enum iATag iATag;
typedef enum iPTag iPTag;
typedef enum iLogicalType iLogicalType;
typedef enum iDAFSWriteMode iDAFSWriteMode;
typedef enum iDAFSCompType iDAFSCompType;
typedef enum iSearchType iSearchType;
typedef enum iMess iMess;
#endif


/* these fellows are hidden in entity.h -- these are the working definitions */
struct iEntity;
struct iEditRecord;
struct iPropMap;
struct iImage;

/* high hackery */
struct  svbmap;
typedef struct svbmap *iPropPtr;

struct iPosSet {
	iUCode chr;  /* The character     */
	unsigned short conf; /* The confidence value  */
} ;


/* could be an ascii or uni expression ? */
struct iRegex {
	uint2 size;  /* size of compiled reg. exp. in bytes */
	void *data; /* compiled reg. exp. */
};

struct iBox {
	short x,y,w,h;
};

/* keep around for older files that still use Long values   */
struct iLBox { 
	int4 x,y,w,h;		 
};

struct iPoint {
    short x, y;
} ;


struct iOrthogon {
	int2 n;
	struct iPoint *point;
};

typedef struct iOrthogon *iOrthogonPtr;
typedef struct iOrthogon **iOrthogonArrayPtr;

struct iOrthogonList {
	int2 num_poly;
	iOrthogonArrayPtr poly;
};

struct BOX {
    int2 x1, x2, y1, y2;
};

typedef struct BOX BOX, *BoxPtr;

struct iRegionType {
    int2 size;
    int2 numRects;
    BOX *rects;
    BOX extents;
};

/* for iPUserPtr to free thier data */
typedef void (APIENTRY *iUserPtrFree)(void *data);

/* 
// ---------------           messages    -----------------
*/


typedef iDAFSError (APIENTRY *iImageWriteProc)(char *tag,struct iImage *im);
typedef iDAFSError (APIENTRY *iImageReadProc)(char *tag,struct iImage **im);


/* 
// ---------------     collected typedefs    -----------------
// definitions follow below
// "typedef struct x x;" happens automatically in C++
*/
#ifndef __cplusplus
typedef struct iEntity iEntity;
typedef struct iRegionType iRegionType;
typedef struct iOrthogonList iOrthogonList;
typedef struct iEditRecord iEditRecord;
typedef struct iPropMap iPropMap;
typedef struct iRegex iRegex;
typedef struct iPosSet iPosSet;

typedef struct iPData iPData;
typedef struct iBox iBox;
typedef struct iLBox iLBox;
typedef struct iPoint iPoint;
typedef struct iOrthogon iOrthogon;
#endif

/* The types of messages used by i_SetCallBack(). */
typedef iDAFSError (APIENTRY *iCallBackProc)(iMess m,void *v, struct iEntity *e);

typedef struct iEntity *iEntityPtr;
typedef struct iRegionType *iBoxListPtr;
typedef struct iOrthogonList *iOrthogonListPtr;
typedef struct iEditRecord *iEditRecordPtr;
typedef struct iRegex *iRegexPtr;
typedef struct iPosSet *iPosSetPtr;
typedef struct iPData *iPDataPtr;



/* 
// ---------------     tokens    -----------------
*/

#define iZeroSize 0
#define iOneSize 0x1
#define iTwoSize 0x2
#define iFourSize 0x3
#define iSMask 0x3
#define iTMask 0xfc

#define NullTok 0


#define BeginEntity (0x1 << 2)
#define EndEntity (0x2 << 2)
#define Box (0x3 << 2)
#define ImageID (0x4 << 2)
#define BeginImage (0x5 << 2)
#define ByteOrder (0x6 << 2)
#define BeginProp (0x7 << 2)
#define EndProp (0x8 << 2)
#define Data (0x9 << 2)
#define Long (0xa << 2)
#define BTrue (0xb << 2)
#define BFalse (0xc << 2)
#define BorrowEntity (0xe << 2)
#define DPosSet (0xf << 2)
#define D1PosSet (0x10 << 2)
#define CCITT4Chunk (0x11 << 2)
#define EndImage (0x12 << 2)
#define Short (0x13 << 2)
#define Byte (0x14 << 2)
#define ShortNShort (0x15 << 2)
#define LongArray (0x16 << 2)
#define IData (0x17 << 2)
#define ShortArray (0x18 << 2)
#define Poly (0x19 << 2)
#define SBox (0x20 << 2)

#define UnknownToken1 (0x21 << 2)
#define UnknownToken2 (0xec)

#define AString (0xf0)
#define BString (0xf4)
#define TString (0xf8) 
#define DString (0xfc)


/*
  iZeroSize
  iOneSize
  iTwoSize
  iFourSize
  
  These are OR'd into the tokens (see below) which have
  been left shifted by two to make room.  They represent
  the size of the number (in bytes) which holds the size of
  the data.  Thus, if the data is 4,524 bytes (for example
  in an image chunck), it takes two bytes to represent the
  number 4,524 and iTwoSize will be OR'd into the apprpriate
  token (for example, CCITT4Chunk).  The token will then
  be followed by the two bytes whose (short) value is 4,524.
  If the token represents some data structure other than
  straight bytes, the size byte(s) still say how many
  BYTES the data occupies, not, for example, how many
  elements there might be in an array.
  164
  DAFS-B is so designed that even if a program does not
  understand a token, it can skip that token without
  causing a parsing disaster.  Thus even if a token
  (see below) is not understood, the lowest-order two bits
  are understandable.
  
  iSMask
  iTMask
  These two masks when OR'd with the token byte yield the
  size field (iSMask) or the token (iTMask)
  
  NullTok
  "NullTok" is the token placed twice at the start of a
  DAFS-B document which signifies the document is DAFS-B.
  
  BeginEntity
  This token is required at the beginning of any entity.
  
  EndEntity
  This token appears at the end of each entity.
  In this way DAFS is different from SGML where an end
  tag is only required where the next element is of a
  different kind than this element.  In DAFS, an EndEntity
  is always required.
  
  Box
  This token is used to describe a bounding box
  The data will be four ints, the left x, the upper y,
  The width, and the height.
  
  ImageID
  This token tells, for each entity which has an
  image, which document image is to be used.  The ID which
  follows has to match one of the images specified with a
  BeginImage token.  "BeginImage" and "ImageID" are in some
  since mutually exclusive.  "BeginImage" is used the first
  time an image is used.  "ImageID" is used when that image
  is referred to by other entities.  Thus, assume we have
  a document containing an image.  There are several glyphs
  which are child entities of the document and each glyph
  has an image. "BeginImage" is used to specify the image
  in the "document" entity and "ImageID" is used to refer
  to the image in the "glyph" entities.  An example of this
  is given below.
  
  If one of the glyphs then goes into a file all its own,
  the image will be speficied by "BeginImage".  The
  ILLUMINATOR tools handle this properly when extracting
  an entity from a file and putting it in another.
  
  BeginImage
  This token appears before an image
  must be followed by a Box token and specification of the
  image's bounding box, by a token giving the image format,
  by the image data itself, and by an EndImage token.  See
  also "ImageID".
  
  BeginProp
  This token marks the beginning of a collection of property of an
  entity.  It is optional in that entities do not have
  to have declared properties, but it is required wherever
  they do.  Properties are such things as Bold, Italic,
  and Underline, as well as such things as user data.
  
  EndProp
  This token indicates the end of a collection of properties.
  
  Data
  Introduces data whose format is unknown.  Other data
  types which are currently defined are Int, BTrue, and BFalse
  and .  If the data is a single int, or a single Boolean
  use those instead.  For all other data types, use "Data".
  
  When "Data" is declared, the data is presented in the
  file byte by byte.  ILLUMINATOR and its associated tools
  will read the data in and point to it as a (void *).  A
  program which knows the actual data format will recast
  that pointed into whatever it is.  This recasting may be
  to anything: an array of ints, a text string, or a
  complex structure.
  
  The lowest two bits of the "Data" token will be OR'd with
  iOneSize, iTwoSize, or iFourSize.  The following one,
  two, or four bytes will tell the size of the following
  data, and the data will follow immediately after the size
  byte(s).
  
  Long
  This token introduces data which is a single int.  It
  will always be OR'd with iOneSize and be followed by a
  size byte of '4' (the size of one int).  If you want to
  have an array of ints, use the "Data" instead of the "Long"
  token.
  
  BTrue
  BFalse
  These token says the data is a single Boolean, and that the
  value of that Boolean is TRUE.  The last two bits will
  always be 0 for a BTrue or a BFalse, indicating data of
  size zero.  There is no need for data to follow a BTrue
  or a BFalse token, since the value of the token itself
  is the data.
  
  BTrue and BFalse tokens have two unrelated purposes,
  depending on where they are used.  When they are part
  of a property, they indicate the data type is a Boolean
  and what its value is.  When they appear in an entity
  but not between a BeginProp/EndProp pair, they mean that
  the children of this entity are to be read sequentially
  (for BTrue) or as alternatives (for BFalse).
  
  BorrowEntity
  i_ShareEntity() enables us to take an entity and
  effectively put a copy of it where we want to.  What is
  put in the new place is the "BorrowEntity" token and
  data which says from which entity the Shareing took
  place.  This Shareing can only be done when the original
  entity and the Shareed copy have a common ancestor.
  
  Doing it this way (instead of actually copying the entity)
  enables the system to keep track of such Shareed entities
  so that when one copy is changed, the other is also.
  
  DPosSet
  The DPosSet token says what follows is a possibility set
  which can contain any number of character/confidence
  pairs.  The token will usually be OR'd with iOneSize and
  the following size byte will say how many BYTES the
  series of character/confidence pairs occupies.  It is
  NOT the number of such pairs.  An example of a DPosSet
  is given in one of the documents below.
  
  D1PosSet
  This optional token says the data which follows is a
  single possibility set.  It will be OR'd with iOneSize
  and followed by a size byte '4', since a possibility
  set is a structure with a short for the character and
  a short for the confidence.  This is currently only used
  in i_GetGlyph() and i_SetGlyph() where it is known there
  is only one possible character name per glyph.
  
  CCITT4Chunk
  This token says the following data is in CCITT group 4
  format.
  
  EndImage
  This token finishes an image.
  
  ByteOrder
  This is OR'd with iOneSize and followed by the byte '2'
  which says there are two bytes of data following.  Those
  two bytes are written as 0x1.  On a processor which has
  the same byte order as the one on which the document was
  written, this will be read in as 0x1.  On a processor with
  the bytes swapped, it will be read as 0x100 and the
  program will know to swap the bytes in shorts, longs,
  doubles, pointers, and the like.
  
  DAFS is designed so that a program will do something
  sensible even if it cannot interpret a token.  This
  token is the one exception to that.  If this token
  cannot be interpreted by a program, the program will
  crash when it gets a file created on a machine whose
  bytes are in the other order.
  
  
  UnknownToken1
  UnknownToken2
  Tokens with values from UnknownToken1 to UnknownToken2
  (inclusive) are reserved for future DAFS-B expansion.
  
  
  AString
  BString
  TString
  DString
  These four tokens all define ascii strings.  There are
  more than one only to help in parsing an entity.  The
  TString is the entity's type and the DString is the
  entity's contents.
  */

#endif
