#ifndef _SET_
#define _SET_
#include "smemory.h"
#include "RTEMisc.h"
#include "Compare.h"
#include "rafassert.h"

#define DEBUG_SET FALSE

// --------- template for a generic set of objects of type T -------

template<class T> class Set {
        T **element_array;
        int size;
private:
        int containsAt(T *, int *, int, int);
        void removeElementNumber(int);
public:
        Set();
        ~Set();
		void debugPrint(char *);
        int isEmpty() { return size == 0; };
        void clear();
        int addElement(T *);
        T * getAndRemoveElement();
        T * getElementNumber(int);
        int unionWith(Set<T> *);
        void intersectWith(Set<T> *);
        void removeIntersection(Set<T> *);
        void removeElement(T *);
        int contains(T *);
        int numElements() { assert(size >= 0); return size; };
        int copyTo(Set<T> *);
        void deleteContents();
};


// ------------- Member function templates for class set ---------------

template<class T> Set<T>::Set()
{
        size = 0;
        element_array = NULL;
}

template<class T> Set<T>::~Set()
{
        if (element_array) { 
		smfree(element_array); 
		element_array = 0; 
	}
}

template<class T> void Set<T>::debugPrint(char *s)
{
	if(this->isEmpty())
		debugf("set %s is empty\n", s);
	else {
		debugf("Contents of set %s\n", s);
		for(int i=0; i<this->size; i++)
			debugf("element %d is %d\n", i, this->getElementNumber(i));
	}
}

template<class T> void Set<T>::clear()
{
	size = 0;
	if (element_array) { smfree(element_array); element_array = 0; }
}

template<class T> int Set<T>::addElement(T *e)
{
	#ifdef SORT_SET_BY_PACKING_ORDER
		return arrayInsert(&element_array, e, &size);
	#else
		return simpleArrayInsert((void ***)&element_array, e, &size);		
	#endif
}

template<class T> T* Set<T>::getAndRemoveElement()
{
	assert(this->isEmpty() == FALSE);
	T *e = this->getElementNumber(this->size-1);
	this->removeElement(e);
	return e;
}

template<class T> int Set<T>::unionWith(Set<T> *e_set)
{
	if(!this->size && !e_set->size)
	  return TRUE;	

	int total_size = this->size + e_set->size;
	T **union_array = (T **) smalloc(sizeof(T *) * total_size);
	if(!union_array)
	  return FALSE;

	T **set1 = this->element_array, 
	  **set2 = e_set->element_array;
	int index1 = 0, index2 = 0, union_size = 0;
	while(index1 < this->size && index2 < e_set->size) {
	  if(set1[index1] == set2[index2]) {
	    union_array[union_size] = set1[index1++];
	    index2++;
	  }
	  else if(set1[index1] < set2[index2])
	    union_array[union_size] = set1[index1++];
	  else
	    union_array[union_size] = set2[index2++];
	  union_size++;
	}
	if(index1 < this->size) {	  
	  memmove(union_array+union_size, set1+index1, sizeof(T *) * (this->size - index1));
	  union_size += this->size - index1;
	}
	else if(index2 < e_set->size) {
	  memmove(union_array+union_size, set2+index2, sizeof(T *) * (e_set->size - index2));
	  union_size += e_set->size - index2;
	}

	T** smaller_array = (T **) smrealloc(union_array, sizeof(T *) * union_size);
	if(!smaller_array) {
	  smfree(union_array);
	  return FALSE;
	}

	if(this->element_array)
	  smfree(this->element_array);
	this->element_array = smaller_array;
	this->size = union_size;

	return TRUE;
}

template<class T> void Set<T>::intersectWith(Set<T> *e_set)
{
        // this = intersection(this, e_set)

	if(DEBUG_SET) {
		debugf("In Set<T>::intersect_with, prior to intersection:\n");
		this->debugPrint("this");
		e_set->debugPrint("e_set");
	}

	if(!this->isEmpty()) {
		if(e_set->isEmpty())
			this->clear();
		else {
                        const int max_size = MIN(this->size, e_set->size);
                        T **new_element_array = (T**) smalloc(max_size * sizeof(T *));
                        if(new_element_array) {
                                int i=0, j=0, k=0;
                                while (i < this->size && j < e_set->size) {
                                        if (this->element_array[i] < e_set->element_array[j])
                                                i++;
                                        else if (this->element_array[i] > e_set->element_array[j])
                                                j++;
                                        else {
                                                // found a member of the intersection
                                                new_element_array[k++] = this->element_array[i];
                                                i++;  j++;
                                        }
                                }
				if(this->element_array) {
                                	smfree(this->element_array);  
					this->element_array = 0;
				}
                                this->size = k;
                                if(k > 0)
                                        if(k < max_size)
                                                this->element_array = (T**) smrealloc(new_element_array, k*sizeof(T*));
                                        else
                                                this->element_array = new_element_array;
                                else 
                                        smfree(new_element_array);
                        }
		}
	}

	if(DEBUG_SET) {
		debugf("After intersection:\n");
		this->debugPrint("this");
	}
}

template<class T> void Set<T>::removeIntersection(Set<T> *e_set)
{
	// this = this - intersection(this, e_set)
		   
  if(e_set) {
    const int size = e_set->size;
    for(int i=0; i<size; i++)
      this->removeElement(e_set->getElementNumber(i));
  }
}

template<class T> void Set<T>::removeElement(T *e)
{
	// OK to call this rotuine when e is not in the set -
	// otherwise caller would have to perform two bSearches
	// to delete an element which actually is in the set.

	const int start=0, stop=this->size-1;
	int loc, found = this->containsAt(e, &loc, start, stop);
	if(found) {
		assert(compare(this->element_array[loc],e) != LESS_THAN &&
		       compare(this->element_array[loc],e) != GREATER_THAN);

		// make sure we really delete e, not some indistinguishable element
		while (loc < size && this->element_array[loc] != e &&
		       compare(this->element_array[loc], e) != GREATER_THAN) {
			if(DEBUG_SET) printf("remove_element::incrementing loc\n");
			++loc;
		}
		if(loc < size && this->element_array[loc] == e) {
			if(DEBUG_SET) printf("remove_element:: removing element number %d\n", loc);
			this->removeElementNumber(loc);
		} else halt("remove_element called for element not in the set\n");
	}

	assert(this->contains(e) == FALSE);
}

template<class T> void Set<T>::removeElementNumber(int loc)
{
        assert(loc >= 0 && loc < this->size);
        listDelete((void ***)&this->element_array, loc, &this->size);
}

template<class T> int Set<T>::contains(T *e)
{
	const int start=0, stop=this->size-1;
	int loc, found = this->containsAt(e, &loc, start, stop);
	if(found) {
		// make sure we really found e, not some indistinguishable element
		while(loc < size && this->element_array[loc] != e &&
      	      	      compare(this->element_array[loc], e) == 0)
			++loc;
		found = (loc < size && this->element_array[loc] == e);
	}

        assert(!found || (this->element_array[loc] == e));
        return found;
}

template<class T> int Set<T>::containsAt(T *e, int *loc, int start, int stop)
{
        int found=FALSE;
        if(this->size > 0) {
		assert(start >= 0 && start < this->size && stop < this->size && stop >= start);
#ifdef SORT_SET_BY_PACKING_ORDER		
                int lloc = bSearch(&this->element_array[start], e, stop-start+1, &found);
#else
		int lloc = simpleBSearch((void **)&this->element_array[start], e, stop-start+1, &found);
#endif
		if(found) {
			// loop through any indistinguishable elements
			while(lloc <= stop && this->element_array[lloc] != e &&
		      	      compare(this->element_array[lloc], e) == 0)
				++lloc;
			found = (lloc <= stop && this->element_array[lloc] == e);
			if(found) *loc = lloc;
		}
	}
        assert(!found || (this->element_array[*loc] == e));
        return found;
}

template<class T> T * Set<T>::getElementNumber(int i)
{
	if ((i<0) || (i>=this->size))
		halt("index out of range");
	return this->element_array[i];
}

template<class T> int Set<T>::copyTo(Set<T> *e_set)
{
	// e_set becomes a copy of me
	
	int success = TRUE + 1;  // bogus logical value

	if(this->size == 0) {
		// e_set becomes null set
		assert(this->element_array == NULL);
		if(e_set->element_array) {
			assert(e_set->size > 0);
			smfree(e_set->element_array); e_set->element_array = 0;
			e_set->size = 0;
			e_set->element_array = NULL;
		} else {
			assert(e_set->size == 0);
			e_set->size = 0;
		}
		success = TRUE;
	} else {
		const int my_array_size = this->size * sizeof(T *);
		if(e_set->element_array) {
			assert(e_set->size > 0);
			smfree(e_set->element_array); e_set->element_array = 0;
		}
		e_set->element_array = (T **) smalloc(my_array_size);
		if (e_set->element_array) {
			e_set->size = this->size;
			memcpy(e_set->element_array, this->element_array, my_array_size);
			success = TRUE;
		} else {
			// e_set was destroyed as intended, but it becomes
			// a null set instead of a copy of me.
			e_set->size = 0;
			success = FALSE;
		}
	}
	assert((success == TRUE) || (success == FALSE)); // make sure it was set
	return success;
}

template<class T> void Set<T>::deleteContents()
{
	if(this->size > 0) {
		for(int i=0; i<size; i++) {
			delete element_array[i]; 
			element_array[i] = 0;
		}
		smfree(element_array);
		element_array = 0;
		this->size = 0;
	}
}

// names for commonly used set classes
class Rectangle;
class ChildRectangle;
typedef Set<Rectangle> RectangleSet;
typedef Set<ChildRectangle> ElementSet;

#endif
