
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "rafutil.h"

#include "Xm/Xm.h"

#include "smemory.h"
#include "dafs.h"
#include "utf.h"
#include "uString.h"
#include "misc/Unicode.h"

#include "Line.h"
#include "LineList.h"
#include "utf.h"

#define LINE_ALLOC 10

#define FALSE 0
#define TRUE  1

Line::Line(LineList *lines)
{
  alloc_refs = LINE_ALLOC;

  refs = (Reference *) smalloc(alloc_refs * sizeof(Reference));
  num_refs = 0;

  manager = lines;

  indices = (iUCode *) smalloc(2 * sizeof(iUCode));
  indices[0] = indices[1] = 0;
}

Line::Line(const Line &other)
{
  alloc_refs = other.alloc_refs;
  
  refs = (Reference *) smalloc(alloc_refs * sizeof(Reference));

  text    = other.text;
  context = other.context;
  indices = (iUCode *) smalloc((context.length()) + 1 * sizeof(iUCode));
  ucsncpy(indices, other.indices, context.length());
  
  num_refs = other.num_refs;
  for(int  i = 0; i < num_refs; i++) 
    refs[i] = other.refs[i];
}

Line::~Line(void)
{
  smfree(refs);
  if(indices)
    smfree(indices);
}

int
direction(iEntityPtr en, int index)
{  
  return dLEFT;
}

uDirection
Line::direction(int base)
{
  if(text.length() >= 1) {
    switch(uniOrder(text[base2text(base)])) {
    case dLEFT:
      return dLEFT;

    case dRIGHT:
      return dRIGHT;

    default: {
	int pos, prev, next;
	
	pos = base2context(base);
	if(base) {
	  prev = base2context(base-1);
	  if(prev == pos-1)
	    return dLEFT;
	  if(prev == pos+1)
	    return dRIGHT;
	}
	
	if(base <= context.length()) {
	  next = base2context(base+1);
	  if(next == pos+1)
	    return dLEFT;
	  if(next == pos-1)
	    return dRIGHT;
	}
      }
    }
  }

  return dNONE;
}

void
Line::parseContext(void)
{
  if(indices != NULL)
    smfree(indices);
  context = uniContext(text);
  // MAV 990113 -- compile complains that context.string() is const
  // I'm sure this is bad because I know that uniDirection DOES modify input
  uniDirection(context.string(), &indices);
}

void
Line::setText(const iUCode *uni)
{
 text = uni;

 parseContext();
} 

void 
Line::remove(int start, int length)
{
  for(int i = start + length - 1; i >= start; i--)
    removeChar(i);
  
  parseContext();
}

#define dtNONE    0
#define dtNEWLINE 1
#define dtSPACE   2
#define dtUSTRING 3
#define dtSTRING  4
#define dtPOSSET  5

static int
getDataType(iEntityPtr en)
{
  switch(i_GetEntityType(en)) {
  case iNewLineT:
    return dtNEWLINE;
  case iSpaceT:
    return dtSPACE;
  default:
    switch(i_GetEDataType(en)) {
    case iDString:
      return dtSTRING;
    case iDUString:
      return dtUSTRING;
    case iDPosSet:
    case iD1PosSet:
      return dtPOSSET;
    default:
      return dtNONE;
    }
  }
}

static iEntityPtr
dataChild(iEntityPtr en)
{
  switch(i_GetEDataType(en)) {
  case iDUString:
  case iDString:
  case iDPosSet:
  case iD1PosSet:
    return en;
  default:
    if(i_CountChildren(en)) {
      iEntityPtr child, data;
      for(child = i_GetChildEntity(en); child; child = i_GetNextEntity(child))
	if(data = dataChild(child))
	  return data;
    }
    return NULL;
  }
}

static void
cloneStub(iEntityPtr start, iEntityPtr finish, iEntityPtr *en)
{
  if(start != finish) {
    iEntityPtr parent = i_NewEntity(i_GetEntityType(i_GetParentEntity(finish)), 0);
    i_MoveToFront(*en, parent);
    *en = parent;
    
    cloneStub(start, i_GetParentEntity(finish), en);
  }
}

static iEntityPtr 
clone(iEntityPtr start, iEntityPtr finish)
{
  iEntityPtr en = i_NewEntity(i_GetEntityType(finish), 0);

  cloneStub(start, finish, &en);

  return en;
}

static void
cloneSibling(iEntityPtr locus, iEntityPtr *attach, iEntityPtr *data)
{

  *attach = *data = NULL;

  iEntityPtr parent = i_GetParentEntity(locus);
  if(parent && i_CountChildren(parent) > 1) {
    iEntityPtr child;
    for(child = i_GetChildEntity(parent); child; child = i_GetNextEntity(child)) {
      if(child != locus && (*data = dataChild(child))) {
	*attach = clone(child, *data);
	break;
      }
    }
  }

  if(*attach) {
    *data = *attach;
    while(i_GetChildEntity(*data))
      *data = i_GetChildEntity(*data);      
  }
  else  // default case, no entities to match
    *attach = *data = i_NewEntity(iGlyphT, 0);
}

static iEntityPtr
killEntity(iEntityPtr en)
{
  iEntityPtr top, parent = i_GetParentEntity(en);

  if(!i_CountChildren(en) && parent) {
    i_DisposeEntity(en);
    top = killEntity(parent);
  }
  else
    top = en;
  
  return top;
}

void
Line::removeChar(int index)
{
  if(num_refs) {
    int ref = reference(index),
        off = offset(index);
    iEntityPtr root = 0,
               en = refs[ref].en;

    // update references
    growReference(ref, -1);
    text.remove(index);
    if(referenceLength(ref) == 0)
      deleteReference(ref);      

    // update DAFS
    int length;
    switch(getDataType(en)) {
    case dtSTRING:
      length = strlen(i_GetText(en, 0));
      if(length > 1 && off < length) {
	const char *old_string = i_GetText(en, 0);
	char *new_string = (char *) smalloc(length + 1);
	
	strncpy(new_string, old_string, off);
	new_string[off] = '\0';
	strcat(new_string+off, old_string+off+1);

	i_SetText(en, new_string, 0);
	smfree(new_string);
      }
      else if(off == 0)
	root = killEntity(en);
      break;

    case dtUSTRING:
      length = ucslen(i_GetUText(en, 0));
      if(length > 1 && off < length) {
	const iUCode *old_string = i_GetUText(en, 0);
	iUCode *new_string = (iUCode *) smalloc((length + 1) * sizeof(iUCode));
	
	ucsncpy(new_string, old_string, off);
	new_string[off] = 0;
	ucscat(new_string+off, old_string+off+1);

	i_SetUText(en, new_string, 0);
	smfree(new_string);
      }
      else if(off == 0)
	root = killEntity(en);
      break;

    case dtNEWLINE:
    case dtSPACE:
      length = i_GetShort(en, 0);
      if(length > 1 && off < length) 
	i_SetShort(en, length - 1, 0);
      else if(off == 0)
	root = killEntity(en);
      break;

    case dtPOSSET:
      if(off == 0)
	root = killEntity(en);
    }

    // deleted everything on the line
    if(!num_refs) {
      Line *prev = manager->prevLine(this);
      if(prev){
	iEntityPtr last = prev->refs[prev->num_refs-1].en;
	int last_index = prev->refs[prev->num_refs-1].en_index + 1;

	insertReference(0, last, last_index, 0);
      }
      else {
	iEntityPtr new_en = i_NewEntity(iNewLineT, root);
	i_SetShort(new_en, 1, 0);

	insertReference(0, new_en, 0, 0);
      }
    }
  }
  else
    text.remove(index);
}

void
Line::insert(iUCode uc, int index)
{
	const iUCode *old_string = 0;
  // get reference and offset before updating text
  if(num_refs) {
    int ref = reference(index),
        off = offset(index);

	// bogus -- special insert before newline case
	if(index > 0) {
	  int pref = reference(index - 1);
	  if(pref != ref && refs[pref].en == refs[ref].en)
		ref = pref;
	}

    // update dafs
    iEntityPtr en = refs[ref].en;
    
    // want to add data to what is previous to the newline
    if(i_GetEntityType(en) == iNewLineT && num_refs > 1) {
      ref--;
      off = offset(index-1) + 1;
      en = refs[ref].en;
    }

    switch(getDataType(en)) {
    case dtSTRING:
      growReference(ref, 1);
      if(isuascii(uc)) {
		const char *old_string = i_GetText(en, 0);
		char *new_string = (char *) smalloc(strlen(old_string) + 2);
	
		strncpy(new_string, old_string, off);
		new_string[off] = (char) uc;
		new_string[off+1] = '\0';
		strcat(new_string, old_string+off);
	
		i_SetText(en, new_string, 0);
		smfree(new_string);
      }
      else {
		const char *ascii_string = i_GetText(en, 0);
		int length = strlen(ascii_string);
		iUCode *old_string = (iUCode *) smalloc((length + 1) * sizeof(iUCode)),
		  *new_string = (iUCode *) smalloc((length + 2) * sizeof(iUCode));
		utf2ucs(old_string, ascii_string);
	
		ucsncpy(new_string, old_string, off);
		new_string[off] = uc;
		new_string[off+1] = 0;
		ucscat(new_string, old_string+off);
		smfree(old_string);
	
		i_SetUText(en, new_string, 0);
		smfree(new_string);
      }
      break;
      
    case dtUSTRING:
		{
      growReference(ref, 1);

      old_string = i_GetUText(en, 0);
      iUCode *new_string = (iUCode *) smalloc((ucslen(old_string) + 2) * sizeof(iUCode));
      
      ucsncpy(new_string, old_string, off);
      new_string[off] = uc;
      new_string[off+1] = 0;
      ucscat(new_string, old_string+off);
      
      i_SetUText(en, new_string, 0);
		}
      break;
      
    case dtNEWLINE:
      if(uc == '\n') {
		i_SetShort(en, i_GetShort(en, 0) + 1, 0);
		growReference(ref, 1);
		break;
      }
      if(off) {
		iEntityPtr new_en = i_NewEntity(iNewLineT, 0);
		i_SetShort(new_en, i_GetShort(en, 0)-off, 0);
		i_SetShort(en, off, 0);
		i_MoveAfter(new_en, en);
		if(ref >= 0 && ref < num_refs) 
		  changeReference(ref, new_en, -off);
      }

      iEntityPtr attach, data;
      cloneSibling(en, &attach, &data);
      i_SetGlyph(data, uc, 255);
      if (off) {
		i_MoveAfter(attach, en);
		insertReference(0, data, 0, 1);
      }
      else {
		i_MoveBefore(attach, en);
		insertReference(0, data, 0, 1);
      }

      break;

    case dtSPACE:
      if(uc == ' ') {
		i_SetShort(en, i_GetShort(en, 0) + 1, 0);
		growReference(ref, 1);
		break;
      }
      else {
		if(off) {
		  iEntityPtr new_en = i_NewEntity(iSpaceT, 0);
		  i_SetShort(new_en, i_GetShort(en, 0)-off, 0);
		  i_SetShort(en, off, 0);
		  i_MoveAfter(new_en, en);
		  splitReference(ref, index);
		  refs[ref+1].en = new_en;
		  refs[ref+1].en_index = 0;	
		}	
	
		iEntityPtr attach, data;
		cloneSibling(en, &attach, &data);
		i_SetGlyph(data, uc, 255);
		if (off) {
		  i_MoveAfter(attach, en);
		  insertReference(ref+1, data, 0, 1);
		} 
		else {
		  i_MoveBefore(attach, en);
		  insertReference(ref, data, 0, 1);
		}
		break;
      }

    case dtPOSSET:
      iEntityPtr new_en = i_NewEntity(i_GetEntityType(en), NULL);
      
      i_SetGlyph(new_en, uc, 255);
      if(off) {
		i_MoveAfter(new_en, en);
		insertReference(ref+1, new_en, 0, 1);
      }
      else {
		i_MoveBefore(new_en, en);
		insertReference(ref, new_en, 0, 1);
      }
      break;
    }
  }

  // update text
  text.move(index, 1);
  text[index] = uc;
    
  // update context and indices
  parseContext();

}

int
Line::text2base(int raw) 
{
  unsigned base = 0;
  
  raw++;
  while(raw)
    base += uniSpacing(text[raw--]);
  
  return base - 1;
}

int
Line::base2text(int base)
{
	unsigned raw = 0;

	base++;
	while(base > 0)
		base -= uniSpacing(text[raw++]);

	return raw - 1;
}  

int 
Line::context2base(int pos)
{
  if(pos > context.length())
    return indices[context.length()];

  return indices[pos];
}

int 
Line::base2context(int base)
{
  int i, pos = 0;

  for(i = 0; i < context.length(); i++) {
    if(indices[i] == base) 
      return pos;
    if(uniSpacing(context[i]))
      pos++;      
  }

  return pos;
}

int
Line::screen2base(int pos)
{
  if(pos > baseLength())
    return indices[context.length()];

  int i, c = 0;
  for(i = 0; i < pos && c < context.length(); i++) {
    while(indices[c] == indices[c+1]) 
      c++;
    c++;
  }

  return indices[c];
}

int 
Line::base2screen(int base)
{
  int i = 0, c = 0;

  while(c < context.length() && indices[c] != base) {
    if(indices[c] != indices[++c])
      i++;
  }

  if(c == context.length()) {
    i = 0;
    c = 0;
    base++;

    while(c < context.length() && indices[c] != base) {
      if(indices[c] != indices[++c])
	i++;
    }
  }    

  return i;
}  

int
Line::nonspacing(unsigned raw)
{
	int non_spacing = 0;

	while(!uniSpacing(text[raw+non_spacing]))
		non_spacing++;

	return non_spacing;
}

// assumes that the line has a new at the end
Line &
Line::operator +=(Line &other)
{
  // removes the new line '\n'
  if(num_refs)
    removeChar(text.length());

  int text_end = text.length();

  text += other.text;
  
  alloc_refs += other.alloc_refs;
  refs = (Reference *) smrealloc(refs, alloc_refs * sizeof(Reference));

  for(int i = 0; i < other.num_refs; i++) {
    refs[i+num_refs] = other.refs[i];
    refs[i+num_refs].text_index = text_end;
    text_end += other.referenceLength(i);
  }

  num_refs += other.num_refs;

  parseContext();

  // removes the line on the RHS
  manager->detachLine(&other);
}

// adds unicode from a single entity
void
Line::addUnicode(const iUCode *uni, const iEntityPtr en, int index)
{
  insertReference(num_refs, en, index, ucslen(uni));
  text += uni;
}

int 
Line::reference(int index)
{
  int i = num_refs - 1;
  while(i >= 0 && index < refs[i].text_index)
    i--;

  // returns a signal value of -1 if an empty line
  return i;
}

int
Line::offset(int index)
{
  int ref = reference(index);
  
  return refs[ref].en_index + index - refs[ref].text_index;
}

// returns the index of the character at the end of the reference
int 
Line::referenceEnd(int ref)
{
  if(ref < 0)
    return -1;
  else
    return refs[ref].text_index + referenceLength(ref) - 1;
}

int
Line::referenceLength(int ref)
{
  int next;
	int l = text.length();
  if(ref == num_refs-1) {
    next = l;
	// here we see how mr hayner's choice of making the newline
	// a special case not stored in the text string has screwed us
	// again
	// this is off by one when inserting a newline

	if(l && (this->text[l - 1] == '\n'))
	  next++;
  }
  else
    next = refs[ref+1].text_index;

  return next - refs[ref].text_index;
}  

void
Line::splitReference(int reference, int index) 
{
  if(index > refs[reference].text_index && 
     index < refs[reference].text_index + referenceLength(reference)) {

    memmove(refs+reference+2, refs+reference+1, (num_refs-reference) * sizeof(Reference));
    
    refs[reference+1].en = refs[reference].en;
    refs[reference+1].text_index = index;
    refs[reference+1].en_index = refs[reference].en_index + index - refs[reference].text_index;
    
    num_refs++;
    if(num_refs == alloc_refs) {
      alloc_refs += LINE_ALLOC;
      refs = (Reference *) smrealloc(refs, alloc_refs * sizeof(Reference));
    }
  }
}

void
Line::insertReference(int reference, iEntityPtr en, int offset, int length)
{
  memmove(refs+reference+1, refs+reference, (num_refs-reference) * sizeof(Reference));
  
  refs[reference].en = en;
  refs[reference].en_index = offset;
  refs[reference].text_index = referenceEnd(reference-1) + 1;
  
  num_refs++;
  for(int i = reference+1; i < num_refs; i++)
    refs[i].text_index += length;

  if(num_refs == alloc_refs) {
    alloc_refs += LINE_ALLOC;
    refs = (Reference *) smrealloc(refs, alloc_refs * sizeof(Reference));
  }
}

void
Line::growReference(int reference, int amount)
{
  int i;
  Line *line_on;
  iEntityPtr target = refs[reference].en;

  if(reference < num_refs - 1) {
    i = reference + 1;
    line_on = this;
  }
  else {
    i = 0;
    line_on = manager->nextLine(this);
  }
    
  while(line_on && line_on->refs[i].en == target) {
    line_on->refs[i].en_index += amount;
    if(i == line_on->num_refs - 1) {
      line_on = line_on->manager->nextLine(line_on);
      i = 0;
    }
    else
      i++;
  }

  for(i = reference + 1; i < num_refs; i++)
    refs[i].text_index += amount;
}


// used in splitting newlines
void
Line::changeReference(int reference, iEntityPtr en, int amount)
{
  int i;
  Line *line_on;
  iEntityPtr old_en = refs[reference].en;

  refs[reference].en = en;
  refs[reference].en_index += amount;

  if(reference < num_refs - 1) {
    i = reference + 1;
    line_on = this;
  }
  else {
    i = 0;
    line_on = manager->nextLine(this);
  }

  while(line_on && line_on->refs[i].en == old_en) {
    line_on->refs[i].en = en;
    line_on->refs[i].en_index += amount;
    if(i == line_on->num_refs - 1) {
      line_on = manager->nextLine(line_on);
      i = 0;
    }
    else
      i++;
  }
}

void
Line::deleteReference(int reference)
{
  if(num_refs > 1) 
    memmove(refs+reference, refs+reference+1, (num_refs-reference-1) * sizeof(Reference));

  num_refs--;
}

int
Line::calculateLength(Display *display)
{
  render_length = 0;
  
  for(int i = 0; i < this->contextLength() + 1; i++) {
	  iUCode glyph = this->getContext(i);
	  if(uniSpacing(glyph)) {
		  int w, h, a;
		  measureUnicode(display, &glyph, 1, &w, &h, &a);
		  render_length += w;
	  }
  }


  return render_length;
}

// bogus doeznt justify
int
Line::renderLength(Display *display, int x)
{
  int glyph_width;
  glyph_width = 0;

  for(int i = 0; i < x; i++) {
	  iUCode glyph = this->getContext(i);
	  if(uniSpacing(glyph)) {
		  int w, h, a;
		  measureUnicode(display, &glyph, 1, &w, &h, &a);
		  glyph_width += w;
	  }
  }

  return glyph_width;
}

extern void mypr(char *);

void
Line::pr()
{
  iUCode *uc = this->text.string();
  char *ascii = (char *) smalloc(utfMaxSize(uc) * sizeof(char));
  ucs2utf(ascii, uc);
  printf("\t");
  mypr(ascii);
  for(int j = 0; j < this->num_refs; j++) {
	printf("\tfrom the %i char in %p at the %i pos on the line\n", 
		   this->refs[j].en_index, 
		   this->refs[j].en, 
		   this->refs[j].text_index);
  }
}

int
Line::compare(Line *other)
{
  int min_refs;
  int i;
  int fail = FALSE;
  // bogus -- could compare the text but I don't think there
  // will be any difference

  min_refs = MIN(this->num_refs, other->num_refs);
  if(this->num_refs != other->num_refs){
	printf("\tmismatch in the number of refs original:%d other:%d\n",
		   this->num_refs,
		   other->num_refs);
	fail = TRUE;
  }

  for(i = 0; i < min_refs; i++){
	int first_fail = TRUE;
	Reference *my_ref = this->refs + i;
	Reference *other_ref = other->refs + i;
	if(my_ref->en != other_ref->en) {
	  if(first_fail) {
		printf("\tindex %d\n", i);
		first_fail = FALSE;
	  }
	  printf("\t\tmismatch in en original:%p other:%p\n",
			 my_ref->en,
			 other_ref->en);
	}
	if(my_ref->en_index != other_ref->en_index) {
	  if(first_fail) {
		printf("\tindex %d\n", i);
		first_fail = FALSE;
	  }
	  printf("\t\tmismatch in en_index original:%d other:%d\n",
			 my_ref->en_index,
			 other_ref->en_index);
	}
	if(my_ref->text_index != other_ref->text_index) {
	  if(first_fail) {
		printf("\tindex %d\n", i);
		first_fail = FALSE;
	  }
	  printf("\t\tmismatch in text_index original:%d other:%d\n",
			 my_ref->text_index,
			 other_ref->text_index);
	}
	fail = fail || !first_fail;
  }
  
  return !fail;
}
