#include "Rectangle.h"
#include "RTEView.h"
#include "RTEMisc.h"
#include "smemory.h"
#include "stdlib.h"  // for abs()
#include "RTEList.h"
#include "RootElement.h"

#define DEBUG_MISC FALSE

// ------------- Utility functions used outside this module ------------------

Coord smgt(Coord x, Coord k)
{
        // return the smallest multiple of k which is >= x
        if(k == 0 || x == 0)
                return 0;
        else {
                int d = x / k, m = x % k;
                if(m > 0) d++;
                return d * k;
        }
}

Direction opposite(Direction d)
{
        Direction result=RTENone;
        switch (d) {
		    case LeftToRight: result = RightToLeft; break;
		    case RightToLeft: result = LeftToRight; break;
		    case TopToBottom: result = BottomToTop; break;
		    case BottomToTop: result = TopToBottom; break;
		    default: halt("opposite called for direction RTENone"); break;
        }
        return result;
}


int compare(Interval *a, Interval *b)
{
	return b->compare(a);
}

int compare(ChildRectangle *a, ChildRectangle *b)
{
	if(a == b)
		return EQUAL;
	else {
		ParentRectangle *parent = a->getParent();
		if(parent == NULL)
			parent = b->getParent();
		else
			assert(b->getParent() == NULL || b->getParent() == parent);
		if(parent == NULL)
			halt("neither Rectangle being compared has a parent\n");
	
		return parent->inOrder(a, b);
	}
}

int compare(Rectangle *r1, Rectangle *r2)
{
	if(r1 < r2)
		return LESS_THAN;
	else if(r1 > r2)
		return GREATER_THAN;
	else
		return EQUAL;
}

int compare(Coord *c, Interval *i)
{
	return i->compare(*c);
}

Coord nearestMultiple(Coord x, Coord s)
{
	// return the multiple of s nearest to x

	assert(x >= 0 && s > 0);
	const Coord q = x/s, r1 = q*s, r2 = (q+1)*s, d1 = x - r1, d2 = r2 - x;
	if(d1 < d2)
		return r1;
	else
		return r2;
}

Coord nearestPositiveMultiple(Coord x, Coord s)
{
	// return the positive multiple of s nearest to x

	return MAX(nearestMultiple(x, s), s);
}

Coord nearestMultipleGT(Coord x, Coord s)
{
	// return the multiple of s greater than or equal to x nearest to x

	assert(x >= 0 && s > 0);
	if(x%s == 0)
		return x;
	else
		return (x/s + 1) * s;
}

// ---------------- Methods for class selection_manager ----------------------

void RTESelectionManager::select(ChildRectangle *e)
{
	selected_set.addElement(e);
	e->select();
}

void RTESelectionManager::select(ElementSet *e_set)
{
	selected_set.unionWith(e_set);
	const int stop = e_set->numElements();
	for(int i=0; i<stop; i++) 
		e_set->getElementNumber(i)->select();
}

void RTESelectionManager::deselect(ChildRectangle *e)
{
	selected_set.removeElement(e);
	e->unselect();
}

void RTESelectionManager::deselect(ElementSet *e_set)
{
	selected_set.removeIntersection(e_set);
	const int stop = e_set->numElements();
	for(int i=0; i<stop; i++) 
		e_set->getElementNumber(i)->unselect();
}

void RTESelectionManager::toggleSelection(ChildRectangle *e)
{
	if(selected_set.contains(e)) {
		selected_set.removeElement(e);
		e->unselect();
	} else {
		selected_set.addElement(e);
		e->select();
	}
}

void 
RTESelectionManager::toggleSelection(ElementSet *e_set)
{
  int i, size = e_set->numElements();
  for(i = 0; i < size; i++)
    toggleSelection(e_set->getElementNumber(i));
}


void RTESelectionManager::clear()
{
	const int stop = selected_set.numElements();
	for(int i=0; i<stop; i++) 
		selected_set.getElementNumber(i)->unselect();
	first = last = 0;
	selected_set.clear();
}

void RTESelectionManager::getSelected(ElementSet *e_set)
{
	selected_set.copyTo(e_set);
}


int RTESelectionManager::isSelected(ChildRectangle *e)
{
	return selected_set.contains(e);
}

int 
RTESelectionManager::numSelected(void)
{
  return selected_set.numElements();
}

// ---------------- Methods for class TraversalManager --------------------

TraversalManager::TraversalManager()
{
  at_end = FALSE;
}

TraversalManager::~TraversalManager()
{

}

ChildRectangle *
TraversalManager::checkUp(ChildRectangle *child, int goal)
{
	Coord x_pos, y_pos;
	child->getAbsoluteCoords(&x_pos, &y_pos);
	if(goal >= 0)
	  x_pos = goal;
	RootElement *root = child->getRoot();
	Rectangle region(x_pos-root->getX(), 0, child->getW(), y_pos-root->getY());
	ElementSet *set = root->getLeafElementsInRectangle(&region);
	at_end = FALSE;
	
	if(set){
		Coord distance, best_distance = MAXCoord, rect_x, rect_y;
		ChildRectangle *rect, *best_rect = NULL;
		for(int i = 0; i < set->numElements(); i++){
			rect = (ChildRectangle *) set->getElementNumber(i);
			rect->getAbsoluteCoords(&rect_x, &rect_y);
			
			distance = y_pos - (rect_y + rect->getH() - 1);
			if(distance >= 0 && distance < best_distance){
				best_rect = rect;
				best_distance = distance;
			}
		}
		
		delete set;

		if(best_rect && best_rect->numChildren()) {
		  at_end = TRUE;
		  return ((PackedRectangle *) best_rect)->lastChild();
		}
		else
		  return best_rect;
	}
		
	return NULL;
}

ChildRectangle *
TraversalManager::checkDown(ChildRectangle *child, int goal)
{
	Coord x_pos, y_pos;
	child->getAbsoluteCoords(&x_pos, &y_pos);
	if(goal >= 0)
	  x_pos = goal;
	RootElement *root = child->getRoot();
	Rectangle region(x_pos-root->getX(), y_pos+child->getH()-root->getY(), child->getW(), MAXCoord);
	ElementSet *set = root->getLeafElementsInRectangle(&region);
	at_end = FALSE;
	
	if(set){
		Coord distance, best_distance = MAXCoord, rect_x, rect_y;
		ChildRectangle *rect, *best_rect = NULL;
		for(int i = 0; i < set->numElements(); i++){
			rect = (ChildRectangle *) set->getElementNumber(i);
			rect->getAbsoluteCoords(&rect_x, &rect_y);
			
			distance = rect_y - (y_pos + child->getH() - 1);
			if(distance >= 0 && distance < best_distance){
				best_rect = rect;
				best_distance = distance;
			}
		}
		
		delete set;

		if(best_rect && best_rect->numChildren()) {
		  at_end = TRUE;
		  return ((PackedRectangle *) best_rect)->lastChild();
		}
		else
		  return best_rect;
	}
		
	return NULL;
}

ChildRectangle *
TraversalManager::checkLeft(ChildRectangle *child, int goal)
{
	Coord x_pos, y_pos;
	child->getAbsoluteCoords(&x_pos, &y_pos);
	if(goal >= 0)
	  y_pos = goal;
	RootElement *root = child->getRoot();
	Rectangle region(0, y_pos-root->getY(), x_pos-root->getX(), child->getH());
	ElementSet *set = root->getLeafElementsInRectangle(&region);
	at_end = FALSE;
	
	if(set){
		Coord distance, best_distance = MAXCoord, rect_x, rect_y;
		ChildRectangle *rect, *best_rect = NULL;
		for(int i = 0; i < set->numElements(); i++){
		  rect = (ChildRectangle *) set->getElementNumber(i);
		  if(rect->getType() != CursorT) {
		    rect->getAbsoluteCoords(&rect_x, &rect_y);
		    
		    distance = x_pos - (rect_x + rect->getW() - 1);
		    if(distance >= 0 && distance < best_distance){
		      best_rect = rect;
		      best_distance = distance;
		    }
		  }
		}
		
		delete set;
		
		if(best_rect && best_rect->numChildren()) {
		  at_end = TRUE;
		  return ((PackedRectangle *) best_rect)->lastChild();
		}
		else
		  return best_rect;
	}
		
	return NULL;
}

ChildRectangle *
TraversalManager::checkRight(ChildRectangle *child, int goal)
{
	Coord x_pos, y_pos;
	child->getAbsoluteCoords(&x_pos, &y_pos);
	if(goal >= 0)
	  y_pos = goal;
	RootElement *root = child->getRoot();
	Rectangle region(x_pos+child->getW()-root->getX(), y_pos-root->getY(), MAXCoord, child->getH());
	ElementSet *set = root->getLeafElementsInRectangle(&region);
	at_end = FALSE;
	
	if(set){
		Coord distance, best_distance = MAXCoord, rect_x, rect_y;
		ChildRectangle *rect, *best_rect = NULL;
		for(int i = 0; i < set->numElements(); i++){
			rect = (ChildRectangle *) set->getElementNumber(i);
			rect->getAbsoluteCoords(&rect_x, &rect_y);
			
			distance = rect_x - (x_pos + child->getW() - 1);
			if(distance >= 0 && distance < best_distance){
				best_rect = rect;
				best_distance = distance;
			}
		}
		
		delete set;
		
		if(best_rect && best_rect->numChildren()) {
		  at_end = TRUE;
		  return ((PackedRectangle *) best_rect)->lastChild();
		}
		else
		  return best_rect;
	}
		
	return NULL;
}

// -------------------- Methods for class deleted_pool ---------------------

DeletedPoolManager::~DeletedPoolManager()
{
	// make sure all elements in the deleted pool actually get deleted
	while(!this->deleted_pool.isEmpty()) {
		ChildRectangle *e = this->deleted_pool.getAndRemoveElement();
		delete e;  e = 0;
	}
}


void DeletedPoolManager::sendToDeleted(ChildRectangle *e)
{
  // should add it to the deleted_pool set
  delete e;
}

void DeletedPoolManager::sendToDeleted(ElementSet *e_set)
{
	// code TBD
}


void DeletedPoolManager::getFromDeleted(ChildRectangle *e)
{
	// code TBD
}


void DeletedPoolManager::getFromDeleted(ElementSet *e_set)
{
	// code TBD
}


int DeletedPoolManager::isDeleted(ChildRectangle *e)
{
	return deleted_pool.contains(e);
}


// ------------------- Methods for class Interval ------------------------

SimpleInterval::SimpleInterval(Coord px, Coord pdx)
{
	//if(px < 0 || pdx < 0) halt("negative x or dx");
	this->x = px;  this->dx = pdx;
}


Interval::Interval(Coord px, Coord pdx)
	: SimpleInterval(px, pdx)
{
	elements_present = new ElementSet;
}

Interval::Interval(Coord px, Coord pdx, ChildRectangle *r)
	: SimpleInterval(px, pdx)
{
	elements_present = new ElementSet;
	this->elements_present->addElement(r);
}

Interval::~Interval()
{
	delete this->elements_present;
	this->elements_present = 0;
}

void Interval::debugPrint(int index)
{
	debugf("  Interval %d has x = %d, dx = %d\n", index, this->x, this->dx);
	this->elements_present->debugPrint("elements present");
}

void Interval::debugPrint(char *s)
{
	debugf("  %s has x = %d, dx = %d\n", s, this->x, this->dx);
}

int Interval::portionToLeftOf(Interval *i, Interval **r)
{
	// return new Interval containing portion of me to left of i
	Interval *result = NULL;
	int success = TRUE;
	if(this->x < i->x) {
		if(!(result = new Interval(this->x, MIN(i->x - this->x, this->dx))))
			success = FALSE;
		else 
			this->elements_present->copyTo(result->elements_present);
	}
	*r = result;
	return success;
}

int Interval::portionToRightOf(Interval *i, Interval **r)
{
	// return new Interval containing portion of me to right of i
	Interval *result = NULL;
	int success = TRUE;
	if(this->x + this->dx - 1 > i->x + i->dx - 1) {
		if(!(result = new Interval(i->x + i->dx, this->x + this->dx - 1 - (i->x + i->dx - 1))))
			success = FALSE;
		else 
			this->elements_present->copyTo(result->elements_present);
	}
	*r = result;
	return success;

}

int Interval::intersectionWith(Interval *i, Interval **r)
{
	// return new Interval containing intersection of me with i
	Interval *result = NULL;
	int success = TRUE;
	Coord left  = MAX(this->x, i->x),
	      right = MIN(this->x + this->dx - 1, i->x + i->dx - 1);
	if(left <= right) {
		if(!(result = new Interval(left, right-left+1)))
			success = FALSE;
		else {
			this->elements_present->copyTo(result->elements_present);
			result->elements_present->unionWith(i->elements_present);
		}
	}
	*r = result;
	return success;
}

void Interval::removeIntersection(Interval *i)
{
	// remove from me my intersection with i
	if(!i->isNull()) {
		Coord left  = MAX(this->x, i->x),
	      	  right = MIN(this->x + this->dx - 1, i->x + i->dx - 1);
		if(left <= right) {
			// can't remove piece from middle as this would create two Intervals
			if(this->x < left && this->x + this->dx - 1> right)
				halt("Cannot remove section from middle of Interval\n");
			if(this->x < left) {
				this->dx = MAX(0, left - this->x);
			} else {
				this->dx = MAX(0, this->x + this->dx - 1 - right);
				this->x = right + 1;
			}
		}
	}
	assert(this->x >= 0 && this->dx >= 0);
}


int Interval::compare(Coord c)
{
	// position of c relative to me

	if(c < this->x)
		return LESS_THAN;
	else if(c > this->x + this->dx - 1)
		return GREATER_THAN;
	else
		return IS_CONTAINED;
}


int Interval::compare(Interval *i)
{
	// determine position of i relative to me
	
	int result = GREATER_THAN + 1; // initial bogus value

	if(i->x < this->x) 
		if(i->x + i->dx - 1 >= this->x + this->dx - 1)
			result = CONTAINS;
		else
			if(i->x + i->dx - 1 < this->x)
				result = LESS_THAN;
			else
				 result = OVERLAPS;
	else if(i->x == this->x)
		if(i->dx < this->dx)
			result =  IS_CONTAINED;
		else if(i->dx == this->dx)
			result = EQUAL;
		else // (i->dx > this->dx)
			result = CONTAINS;
	else // (i->x > this->x)
		if(i->x + i->dx - 1 <= this->x + this->dx - 1)
			result = IS_CONTAINED;
		else
			if(i->x >  this->x + this->dx - 1)
				result = GREATER_THAN;
			else
				result = OVERLAPS;
	
	assert(result == LESS_THAN || result == IS_CONTAINED || result == CONTAINS ||
	       result == EQUAL || result == GREATER_THAN || result == OVERLAPS);

	return result;
}

ElementSet *Interval::getElementsPresent()
{
	ElementSet *result_set = new ElementSet;
	this->elements_present->copyTo(result_set);
	return result_set;
}


// ----------------- Methods for class IntervalList ------------------------

IntervalList::IntervalList()
{
	members = 0;
	size = 0;
}

IntervalList::~IntervalList()
{
	if(members) {
		assert(size > 0);
		for(int i=0; i<size; i++) {
			delete members[i];
			members[i] = 0;
		}
		smfree(members); members = 0;
	}
}


void IntervalList::debugPrint(char *s)
{
	debugf("Contents of IntervalList %s are:\n", s);
	for(int i=0; i<this->size; i++)
		this->members[i]->debugPrint(i);
}



int IntervalList::insertInterval(Interval *i)
{
	return arrayInsert((void ***) &this->members, i, &this->size, compareII);
}


ElementSet *IntervalList::findHit(Coord c)
{
	ElementSet *result_set = new ElementSet;
	int found, loc = bSearch((void **) members, &c, size, &found, compareCI);
	if(found) 
		members[loc]->elements_present->copyTo(result_set);
	return result_set;
}


ElementSet *IntervalList::findHit(Interval *i)
{
	// return set of all Intervals fully or partially in i
	
	ElementSet *result_set = new ElementSet;
	int found, loc = bSearch((void **) members, i, size, &found, compareII);
	assert(!found || loc == 0 || compare(members[loc-1], i) == LESS_THAN);
	if(found)
		do 
			result_set->unionWith(members[loc]->elements_present);
		while (++loc < size && compare(members[loc], i) != GREATER_THAN );
	return result_set;
}


ElementSet *IntervalList::findIncluded(Interval *i)
{
	// return set of all Intervals fully contained in i

	ElementSet *result_set = new ElementSet;
	int found, loc = bSearch((void **) members, &i->x, size, &found, compareII);
	if(found) {
		int c=compare(members[loc], i);
		do 
			if(c == CONTAINS || c == EQUAL)
				result_set->unionWith(members[loc]->elements_present);
		while (++loc < size && (c=compare(members[loc], i)) != GREATER_THAN);
	}
	return result_set;
}

int IntervalList::noOverlaps()
{
	int i;
	for(i=0; i<this->size-1; i++) {
		if(members[i]->x + members[i]->dx > members[i+1]->x)
			return FALSE;
	}
	return TRUE;
}

int IntervalList::add(Interval **pi)
{
	Interval *i = *pi, *i1, *i2, *i3, *i4;

	if(i->isNull()) halt("called with null Interval");
	if(i->elements_present == NULL) halt("called with no elements present\n");
	if(i->isEmpty()) halt("called with empty Interval");
	if(DEBUG_MISC) i->debugPrint("being added in IntervalList::add");

	if(this->size == 0) {
		if(!listInsert((void ***) &this->members, i, 0, &this->size)) 
		  goto failure;
		//return this->insertInterval(i);
	} else {
		// find largest j s.t. members[j-1] < i;  set to zero if no such j
		int found, j = bSearch((void **)members, &i->x, size, &found, compareII);
		assert(j == 0 || compare(members[j-1], i) == LESS_THAN);
		
		if(j == this->size) {
			// new Interval is entirely to the right of all existing ones
			if(!listInsert((void ***)&this->members, i, j, &this->size)) goto failure;
			//return this->insertInterval(i);
		} else {
			while(!i->isNull()) {
				assert(j < size);
				i1 = i2 = i3 = i4 = NULL;
				if(!i->portionToLeftOf(members[j], &i1)) goto failure;
				if(!i1) {
					if(!members[j]->portionToLeftOf(i, &i1)) goto failure;
					if(i1)
						members[j]->removeIntersection(i1);
				}
				if(!i->intersectionWith(members[j], &i2)) goto failure;
				if(j == size-1) 
					if(!i->portionToRightOf(members[j], &i4)) goto failure;
				members[j]->removeIntersection(i); 
				i3 = members[j];
				listDelete((void ***)&this->members, j, &this->size);
				if(i3->isNull()) {
					delete(i3); i3 = 0;
				}
	
				if(i1) {
					assert(i1->elements_present != NULL);
					if(!listInsert((void ***)&this->members, i1, j, &this->size)) goto failure;
					//if(!this->insertInterval(i1)) goto failure;
					j++;
					assert(this->noOverlaps());
				}
				if(i2) {
					assert(i2->elements_present != NULL);
					if(!listInsert((void ***)&this->members, i2, j, &this->size)) goto failure;
					//if(!this->insertInterval(i2)) goto failure;
					j++;
					assert(this->noOverlaps());
				}
				if(i3) {
					assert(i3->elements_present != NULL);
					if(!listInsert((void ***)&this->members, i3, j, &this->size)) goto failure;
					//if(!this->insertInterval(i3)) goto failure;
					j++;
					assert(this->noOverlaps());
				}
				if(i4) {
					assert(i4->elements_present != NULL);
					if(!listInsert((void ***)&this->members, i4, j, &this->size)) goto failure;
					//if(!this->insertInterval(i4)) goto failure;
					j++;
					assert(this->noOverlaps());
				}
	
				if(i1) i->removeIntersection(i1);
				if(i2) i->removeIntersection(i2);
				if(i4) i->removeIntersection(i4);

			}
			delete i; i = 0;
		}
	}
	*pi = NULL;
	assert(isSorted(members,size));
	assert(this->noOverlaps());
	if(DEBUG_MISC) this->debugPrint("after insertion");
	return TRUE; // success

failure:
	halt("insertion failed\n");
	assert(0 == 1);  // halt if assertions are active
	if(i1) { 
		delete i1; i1 = 0;
	}
	if(i2) {
		delete i2; i2 = 0;
	}
	if(i4) { 
		delete i4; i4 = 0;
	}
	return FALSE;
}


void IntervalList::deleteElement(ChildRectangle *e)
{
	// Warning - this is slow.  Try to use the version below,
	// unless you don't know where e was originally located.

	int i=0;
	while(i < this->size) {
		members[i]->elements_present->removeElement(e);
		if(members[i]->isEmpty()) {
			Interval *i_delete = members[i];
			listDelete((void ***)&this->members, i, &this->size);
			delete(i_delete);
		} else ++i;
	}
	if(DEBUG_MISC) this->debugPrint("after insertion");
}


void IntervalList::deleteElement(Interval *i, ChildRectangle *e)
{
	// Assume Interval i has same bounds as the Interval originally
	// used when inserting e into the list.

	if(DEBUG_MISC) i->debugPrint("being deleted in IntervalList::delete_element");
	int found, loc = bSearch((void **) members, i, size, &found, compareII);
	
	if(found) {
		assert(loc == 0 || compare(&i->x, members[loc-1]) == GREATER_THAN);
		do {
			members[loc]->elements_present->removeElement(e);
			if(members[loc]->isEmpty()) {
				Interval *i_delete = members[loc];
				listDelete((void ***)&this->members, loc, &this->size);
				delete(i_delete);
			}
			else ++loc;
		} while (loc < this->size && compare(i, members[loc]) != LESS_THAN);
	} else
		halt("x location of specified Interval not found");
	if(DEBUG_MISC) this->debugPrint("after insertion");
}


void IntervalList::deleteElements(ElementSet *s)
{
	// Warning - this is slow.  Try to use the version below,
	// unless you don't know where elements of s were originally located.

	int i=0;
	while(i < this->size) {
		members[i]->elements_present->removeIntersection(s);
		if(members[i]->isEmpty()) {
			Interval *i_delete = members[i];
			listDelete((void ***)&this->members, i, &this->size);
			delete(i_delete);
		} else ++i;
	}
}


void IntervalList::deleteElements(Interval *i, ElementSet *s)
{
	// Assume Interval i has same bounds as the Interval originally
	// used when inserting all elements of s into the list.

	int found, loc = bSearch((void **)members, i, size, &found, compareII);
	assert(loc == 0 || compare(&i->x, members[loc-1]) == GREATER_THAN);
	
	if(found) 
		do  {
			members[loc]->elements_present->removeIntersection(s);
			if(members[loc]->isEmpty()) {
				Interval *i_delete = members[loc];
				listDelete((void ***)&this->members, loc, &this->size);
				delete(i_delete);
			}
			else ++loc;
		} while (loc < this->size && compare(i, members[loc]) != LESS_THAN);
	else
		halt("x location of specified Interval not found");
}



int compareII(const void *v1, const void *v2)
{ 
	 Interval *i1 = (Interval *) v1;
	 Interval *i2 = (Interval *) v2;
  return compare(i1, i2); 
}

int compareCRCR(const void *v1, const void *v2)
{ 
  ChildRectangle *cr1 = (ChildRectangle *) v1;
  ChildRectangle *cr2 = (ChildRectangle *) v2;
  
  return compare(cr1, cr2); 
}

int compareRR(const void *v1, const void *v2)
{ 
	Rectangle *r1 = (Rectangle *) v1;
	Rectangle *r2 = (Rectangle *) v2;
  
  return compare(r1, r2); 
}

int compareInterval(const void *v1, const void *v2)
{ 
	Interval *i1 = (Interval *) v1;
	Interval *i2 = (Interval *) v2;
  return compare(i1, i2); 
}

int compareChildRectangle(const void *v1, const void *v2)
{ 
	 ChildRectangle *cr1 = (ChildRectangle *) v1;
	 ChildRectangle *cr2 = (ChildRectangle *) v2;
  
  return compare(cr1, cr2); 
}

int compareRectangle(const void *v1, const void *v2)
{ 
	Rectangle *r1 = (Rectangle *) v1;
	Rectangle *r2 = (Rectangle *) v2;
  
  return compare(r1, r2); 
}

int compareCI(const void *v1, const void *v2)
{ 
	 Coord *c = (Coord *) v1;
	 Interval *i = (Interval *) v2;
  return compare(c, i); 
}
