#include "VDD.h"
#include "RTEView.h"
#include "Rectangle.h"
#include "RTECursor.h"
#include "RTEMisc.h"
#include "RTEList.h"
#include "UCRectangle.h"
#include "RootElement.h"

UCRectangle::UCRectangle(uString &text)
	: ChildRectangle(0, 0, 1, 1)
{	
	glyph = text;
	unselect();
}

UCRectangle::~UCRectangle()
{
	// nothing to do?
}

void
UCRectangle::measure(VirtualDisplayDevice *vdd)
{
  int width, height, ascent;
  vdd->measureGlyph(context[0], &width, &height, &ascent);
  
  Coord w = MAX(1, width),
        h = MAX(1, height);

  a = ascent;  
  if(this->changeSizeTo(w, h))
    this->setPositionValid();
  else {
    RTECursor *c = this->getRoot()->getCursor();
    this->getParent()->deleteChild(this);
    c->requestRedraw();
  }
}

uString UCRectangle::getUString(void)
{
	return glyph;
}

int UCRectangle::draw(void)
{
	// Don't draw rectangles which are not at a valid position
	if(this->positionIsValid()) {
		VirtualDisplayDevice *vdd = getRoot()->getDisplayDevice();

		Coord x_pos, y_pos;
		getAbsoluteCoords(&x_pos, &y_pos);

		if(is_selected()){
		  vdd->fillRect(x_pos, y_pos, w, h);
		  vdd->inverse();
		}

		int i, length = context.length();
		for(i = 0; i < length; i++)
		  vdd->drawGlyph(context[i], x_pos, y_pos, a);

		if(is_selected())
		  vdd->inverse();
	}
	return 0;
}

int UCRectangle::setContext(uString &text)
{
	if(context != text){
		context = text;
		measure(getRoot()->getDisplayDevice());
		return TRUE;
	}

	return FALSE;
}

int UCRectangle::drawRegion(Rectangle *rect)
{
	draw();
	return 0;
}

int UCRectangle::erase()
{
	int result = TRUE;

	VirtualDisplayDevice *vdd = getRoot()->getDisplayDevice();
	if(vdd) {
	  Coord x_pos, y_pos;
	  getAbsoluteCoords(&x_pos, &y_pos);
	  vdd->eraseRect(x_pos, y_pos, w, h);
	}

	return result;  // success
}

void
UCRectangle::addMark(iUCode mark)
{
	glyph += mark;
}


int UCString::line_spacing;

UCString::UCString(KeyMap *km, Coord width) 
{
        spacing = line_spacing;

	setX(0);
	setY(0);
	setW(width);
	setH(line_spacing);

	iUCode letter;
	int direction;
	km->getInfo(&direction, &letter);

	secondary_writing_direction = TopToBottom;
	can_grow_down  = TRUE;
	can_grow_up    = TRUE;
	can_grow_right = TRUE;
	can_grow_left  = TRUE;

	if(direction == NORMAL)
		primary_writing_direction = LeftToRight;
	else
		primary_writing_direction = RightToLeft;

	do_context = TRUE;

}

UCString::UCString(Direction direction)
{
  spacing = line_spacing;

  setX(0);
  setY(0);
  setW(1);
  setH(line_spacing);

  primary_writing_direction = direction;
  secondary_writing_direction = TopToBottom;
  can_grow_down  = TRUE;
  can_grow_up    = TRUE;
  can_grow_right = TRUE;
  can_grow_left  = TRUE;

  do_context = TRUE;
}   

uString
UCString::assemble(int start, int stop)
{
	uString line;
	UCRectangle *child;

	for(int i = start; i <= stop; i++) {
		child = (UCRectangle *) element_list->getItemNumber(i);
		if(child->getType() == UCRectangleT)
			line += child->getUString();
	}

	return line;
}

uString
UCString::assemble(void)
{
	return assemble(0, element_list->getSize()-1);
}

uString
UCString::getString(void)
{
	return assemble();
}   

void
UCString::context(int start, int stop, int index)
{
  if(do_context) {
    uString glyphs = uniContext(assemble(start, stop));
    
    UCRectangle *child;
    int i = 0, j;

    for(j = start; j <= stop; j++) {
      child = (UCRectangle *) element_list->getItemNumber(j);
      
      if(child->getType() == UCRectangleT){
	uString glyph = glyphs[i];
	while(!uniSpacing(glyphs[++i]))
	  glyph += glyphs[i];
	if(j >= index-1 && j <= index+1 && child->setContext(glyph)) {
	  child->erase();
	  child->draw(); // draw the glyph if context has changed
	}
      }
    }
  }
}

void
UCString::context(void)
{
	return context(0, element_list->getSize()-1, 0);
}

void
UCString::performContext(UCRectangle *glyph) 
{
  int start, stop, index = element_list->locationOfItem(glyph);

  // assumes that no more than two glyphs away will affect context
  start = MAX(index - 2, 0);
  stop  = MIN(index + 2, numChildren()-1);

  context(start, stop, index);
}

void
UCString::keyPress(iUCode key_stroke)
{
  RootElement *root = this->getRoot();
  RTECursor *cursor = root->getCursor();

  UCRectangle *ucr;
  int result = FALSE;
  if(uniSpacing(key_stroke)) {
	ucr = new UCRectangle(key_stroke);
	if(key_stroke < 0x0590 || key_stroke > 0x06FF) {
	  VirtualDisplayDevice *vdd = getRoot()->getDisplayDevice();	  
	  int width, height, ascent;
	  vdd->measureGlyph(key_stroke, &width, &height, &ascent);
	  ucr->changeSizeTo(width, height);
	}
	if(element_list->contains(cursor)) {
	  if(insertBefore(cursor, ucr))
		result = TRUE;		
	}			  
	else {
	  if(insertLast(ucr))
		result = TRUE;
	}
  }
  else {		
	ucr = (UCRectangle *) element_list->getItemNumber(element_list->getSize()-1);
	ucr->addMark(key_stroke);
  }

  if(result) {
	int start, stop, index = element_list->locationOfItem(ucr);
	  
	// assumes that no more than two glyphs away will affect the context of the new glyph
	start = MAX(index - 2, 0);
	stop  = MIN(index + 2, element_list->getSize()-1);
	  
	context(start, stop, index);
  }
}

void
UCString::changeDirection(Direction direction)
{
  if(direction != primary_writing_direction) {
    erase();
    primary_writing_direction = direction;

    ElementList *children = getAllChildren();    
    deleteAllChildren();
    for(int i = 0; i < children->getSize(); i++)
      insertLast(children->getItemNumber(i));

    delete children;
  }
}

int UCPara::line_spacing;
    
UCPara::UCPara(Coord width)
{
  spacing = line_spacing;
  primary_writing_direction = LeftToRight;
  secondary_writing_direction = TopToBottom;

  w = width;
  h = spacing;

  can_grow_up     = TRUE;
  can_grow_down   = TRUE;
  can_grow_left   = TRUE;
  can_grow_right  = TRUE;
}

void
UCPara::keyPress(iUCode key_stroke)
{
  RootElement *root = getRoot();
  RTECursor *cursor = root->getCursor();
  UCString *string_on = (UCString *) cursor->getParent();

  UCRectangle *ucr;
  int result = FALSE;
  if(uniSpacing(key_stroke)) {
    ucr = new UCRectangle(key_stroke);

    // precompute size for non-contextually altering characters
    if(key_stroke < 0x0590 || key_stroke > 0x06FF) { 
      VirtualDisplayDevice *vdd = root->getDisplayDevice();
      
      int width, height, ascent;
      vdd->measureGlyph(key_stroke, &width, &height, &ascent);
      ucr->changeSizeTo(width, height);
    }

    if(string_on) {
      if(!string_on->insertBefore(cursor, ucr)) {
	UCString *clone = new UCString(string_on->getPrimaryWritingDir());
	insertAfter(string_on, clone);
	string_on->deleteChild(cursor);
	clone->insertFirst(cursor);
	string_on = clone;
	if(string_on->insertBefore(cursor, ucr))
	  result = TRUE;
      }
      else
	result = TRUE;
    }
    else{
      UCString *last = (UCString *) lastChild();
      if(!last->insertLast(ucr)) {
	UCString *clone = new UCString(last->getPrimaryWritingDir());
	insertLast(clone);
	if(clone->insertLast(ucr))
	  result = TRUE;
      }
      else
	result = TRUE;
    }
  }
  else {
    if(string_on)
      ucr = (UCRectangle *) string_on->prevChild(cursor);
    else
      ucr = (UCRectangle *) ((UCString *) lastChild())->lastChild();

    if(ucr)
      ucr->addMark(key_stroke);
  }
   
  if(result) {
    if(string_on)
      string_on->performContext(ucr);
    else
      ((UCString *) lastChild())->performContext(ucr);
  }
}
    
uString
UCPara::getString(void)
{
  uString string;
  ElementList *children = getAllChildren();
    
  for(int i = 0; i < children->getSize(); i++)
    string += ((UCString *) children->getItemNumber(i))->getString();

  return string;
}

void
UCPara::changeDirection(Direction direction)
{
  UCString *new_string = new UCString(direction);

  RTECursor *cursor = getRoot()->getCursor();
  cursor->getParent()->deleteChild(cursor);
  new_string->insertLast(cursor);

  insertLast(new_string);
}
 
// assumes a structure of para, string, unicode char
int
UCPara::numGlyphs(void) 
{
  int i, glyphs = 0;
  for(i = 0; i < numChildren(); i++)
    glyphs += element_list->getItemNumber(i)->numChildren();

  return glyphs;
}



