
/*
 * Copyright (c) 1994, 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
// Implements the abstract data type DLList
*/

#ifndef DEBUG_VALUE
#ifdef DEBUG
#define DEBUG_VALUE 1
#else
#define DEBUG_VALUE 0
#endif
#endif

#include "rafport.h"
#include "c-lib.h"
#include "rafutil.h"

#include "dllist.h"

static void 
lnCtor(DLListNode *that, VoidPtr input_value)
{
	that->value = input_value;
	that->next = that->prev = 0;
}

static void
lnDtor(DLListNode *that)
{
	smfree(that);
}

static void 
dllistSplice(DLListNode *first, DLListNode *second)
{
	first ->next = second;
	second->prev = first;
}

static void 
dllistSever(DLListNode *node)
{
	node->prev->next = node->next;
	node->next->prev = node->prev;
	
	lnDtor(node);
}

void
dllistCtor(DLList * that)
{
	that->count = 0;
	
	that->start.prev = 0;
	that->start.next = &(that->end);
	that->start.value = 0;
	that->end.prev = &(that->start);
	that->end.next = 0;
	that->end.value = 0;
}

void dllistDtor(DLList *that)
{
	dllistClear(that);
}

dllError 
dllistAssign(DLList *lvalue, const DLList *rvalue)
{
	Ref r;
	dllError e=DLLIST_OK;
	VoidPtr value;
	
	dllistClear(lvalue);
	for(r = dllistFirst(rvalue); r != 0; dllistNext(rvalue, &r)){
		value = dllistContents(rvalue, r);
		if (e=dllistAppend(lvalue, value))
			goto error;
	}
error:
	return e;
}

int 
dllistEmpty(const DLList *that)
{
	return that->count == 0;
}

int4 
dllistLength(const DLList *that)
{
	return that->count;
}

dllError
dllistAppend(DLList *that, VoidPtr value)
{
	DLListNode *node;
	DLListNode *prev = that->end.prev;
	node = (DLListNode *) smalloc(sizeof(DLListNode));
#ifdef CAREFUL
	if(node == 0)
		halt("ran out of memory.\n");
#endif
	if (!node) return DLLIST_MEMERROR;
	lnCtor(node, value);
	dllistSplice(prev, node);
	dllistSplice(node, &(that->end));
	that->count++;
	return DLLIST_OK;
}

dllError
dllistPrepend(DLList *that, VoidPtr value)
{
	DLListNode *node;
	DLListNode *next = that->start.next;
	node = (DLListNode *) smalloc(sizeof(DLListNode));
	if (!node) return DLLIST_MEMERROR;
	lnCtor(node, value);
	dllistSplice(&(that->start), node);
	dllistSplice(node, next);
	that->count++;
	return DLLIST_OK;
}

void 
dllistJoin(DLList *that, DLList *other)
{
	int4 that_empty = dllistEmpty(that);
	int4 other_empty = dllistEmpty(other);
	DLListNode *last_of_that = that->end.prev;
	DLListNode *first_of_other = other->start.next;
	DLListNode *last_of_other = other->end.prev;
	
	if(other_empty)
		; /* do nothing */
	else {
		if(that_empty) {
			dllistSplice(&(that->start), first_of_other);
		} else {
			dllistSplice(last_of_that, first_of_other);
		}
		dllistSplice(&(that->end), last_of_other);
		dllistSplice(&(other->start), &(other->end));
	}
	that->count+=other->count;
}

VoidPtr 
dllistFront(const DLList *that)
{
	if(dllistEmpty(that))
		halt("dllistFront::accessed empty list.\n");
	return that->start.next->value;
}

VoidPtr dllistRear(const DLList *that)
{
	if(dllistEmpty(that))
		halt("dllistRear::accessed empty list.\n");
	return that->end.prev->value;
}

VoidPtr dllistPopFront(DLList *that)
{
	DLListNode *front;
	VoidPtr value;
	
	if(dllistEmpty(that))
		halt("dllistPopFront::accessed empty list.\n");
	
	front = that->start.next;
	value = front->value;
	dllistSever(front);
	that->count--;
	
	return value;
}

void dllistDeleteFront(DLList *that)
{
	DLListNode *front;
	
	if(dllistEmpty(that))
		halt("dllistDeleteFront::accessed empty list.\n");
	
	front = that->start.next;
	dllistSever(front);
	that->count--;
}

void 
dllistClear(DLList *that)
{
	DLListNode *node;
	DLListNode *next_node;
	
	for(node = that->start.next; node != &(that->end); node = next_node){
		next_node = node->next;
		lnDtor(node);
	}
	
	dllistSplice(&(that->start), &(that->end));
	that->count = 0;
}

Ref
dllistInsertAfter(DLList *that, Ref r, VoidPtr item)
{
	DLListNode *node = (DLListNode *) r;
	DLListNode *new_node;
	DLListNode *after_node;
	
	if(node == 0)
		node = &that->start;
	
	after_node = node->next;
	new_node = (DLListNode *) smalloc(sizeof(DLListNode));
	if (!new_node) return NULL;
	lnCtor(new_node, item);
	
	dllistSplice(node, new_node);
	dllistSplice(new_node, after_node);
	that->count++;
	return new_node;
}

void dllistDeleteAfter(DLList *that, Ref r)
{
	DLListNode *node = (DLListNode *) r;
	DLListNode *next_node;
	
	node = node ? node : &(that->start);
	next_node = node->next;
	
	if(next_node == &(that->end))
		halt("dllistDeleteAfter::attempted to delete after the end of list.\n");
	dllistSever(next_node);
	that->count--;
}

/* Doubly linked stuff */


VoidPtr dllistPopRear(DLList *that)
{
	DLListNode *rear;
	VoidPtr value;
	
	if(dllistEmpty(that))
		halt("dllistPopRear::accessed empty list.\n");
	
	rear = that->end.prev;
	value = rear->value;
	dllistSever(rear);
	that->count--;
	
	
	return value;
}

void dllistDeleteRear(DLList *that)
{
	DLListNode *rear;
	
	if(dllistEmpty(that))
		halt("dllistDeleteRear::accessed empty list.\n");
	
	rear = that->end.prev;
	dllistSever(rear);
	that->count--;
}

Ref
dllistInsertBefore(DLList *that, Ref r, VoidPtr item)
{
	DLListNode *node = (DLListNode *) r;
	DLListNode *new_node;
	DLListNode *before_node;
	
	if(node == 0)
		node = &that->end;
	
	before_node = node->prev;
	new_node = (DLListNode *) smalloc(sizeof(DLListNode));
	if (!new_node) return NULL;
	lnCtor(new_node, item);
	
	dllistSplice(before_node, new_node);
	dllistSplice(new_node, node);
	return new_node;
}


void dllistDelete(DLList *that, Ref *r, int4 direction)
{
	DLListNode *node = (DLListNode *) *r;
	DLListNode *new_node;
	
	switch(direction){
    case DLLIST_FORWARD:
		new_node = node->next;
		break;
    default:
    case DLLIST_BACKWARD:
		new_node = node->prev;
		break;
	}
	
	new_node = ((new_node == &that->start) || (new_node == &that->end)) 
		? 0 : new_node;
	*r = (Ref) new_node;
	dllistSever(node);
	that->count--;
}

Ref
dllistFirst(const DLList *that)
{
	return (that->start.next != &(that->end)) ? that->start.next : 0;
}

Ref
dllistLast(const DLList *that)
{
	return (that->end.prev != &(that->start)) ? that->end.prev : 0;
}

void
dllistNext(const DLList *that, Ref* ref)
{
	DLListNode *node = (DLListNode *) *ref;
	*ref = (node->next != &(that->end)) ? node->next : 0;
}

void
dllistPrev(const DLList *that, Ref* ref)
{
	DLListNode *node = (DLListNode *) *ref;
	*ref = (node->prev != &(that->start)) ? node->prev : 0;
}

VoidPtr
dllistContents(const DLList *that, Ref ref) 
{
	DLListNode *node = (DLListNode *) ref;
	return node->value;
}

Ref
dllistSeek(const DLList *that, VoidPtr value)
{
	Ref r;
	VoidPtr v;
	
	for(r = dllistFirst(that); r != 0; dllistNext(that, &r)){
		v = dllistContents(that, r);
		if(v == value)
			return r;
	}
	return 0;
}


void
dllistRemove(DLList *that, Ref r)
{
	DLListNode *node = (DLListNode *) r;
#if DEBUG_VALUE >= 2
	dllistCheck(that);
#endif
	dllistSever(node);
	that->count--;
#if DEBUG_VALUE >= 2
	dllistCheck(that);
#endif
}

int 
dllistEqual(const DLList *left, const DLList *right)
{
	Ref left_r;
	Ref right_r;
	VoidPtr value_r;
	VoidPtr value_l;
	
	if(dllistLength(left) != dllistLength(right))
		return 0;
	
	for(left_r = dllistFirst(left), right_r = dllistFirst(right); left_r; 
																  dllistNext(left, &left_r), dllistNext(right, &right_r)){
		value_l = dllistContents(left, left_r);
		value_r = dllistContents(right, right_r);
		if(value_l != value_r)
			return 0;
	}
	
	return 1;
}

void
quickSort(VoidPtr *array, int4 f, int4 l, ComparisonFunction compare_function);

/*
// bogus should use the version that doesn't use the sentinel value
*/

static ComparisonFunction f;

#define SENTINEL_VALUE ((VoidPtr) -1)

static int4
newCompare(VoidPtr a, VoidPtr b)
{
	if(a == SENTINEL_VALUE)
		return 1;
	if(b == SENTINEL_VALUE)
		return -1;
	return f((void*)a, (void*)b);
}

/*
// Danger -- after a sort -- the refs all change 
*/
dllError 
dllistSort(DLList *that, ComparisonFunction compare_function)
{
	VoidPtr *array;
	int4 n;
	Ref r;
	VoidPtr value;
	int4 i;
	
	n = dllistLength(that);
	array = (VoidPtr *) smalloc(sizeof(VoidPtr) * (n + 1));
	if (!array) return DLLIST_MEMERROR;
	
	for(i = 0, r = dllistFirst(that); i <n && r; i++, dllistNext(that, &r)){
		value = (VoidPtr) dllistContents(that, r);
		array[i] = value;
	}
	array[n] = SENTINEL_VALUE;
	
	f = compare_function;
	quickSort(array, 0, n - 1, newCompare);
	dllistClear(that);
	
	for(i =0; i < n; i++){
		value = array[i];
		dllistAppend(that, value);
	}
	
	smfree(array);
	return DLLIST_OK;
}

void
quickSort(VoidPtr *array, int4 f, int4 l, ComparisonFunction compare_function)
{
	int4 i, j;
	VoidPtr temp;
	
	if( f >= l)
		return;
	
	i = f + 1;
	while(compare_function(array[i], array[f]) < 0)
		i++;
	
	j = l;
	while(compare_function(array[j], array[f]) > 0)
		j--;
	
	while(i < j) {
		temp = array[i];
		array[i] = array[j];
		array[j] = temp;
		i++;
		while(compare_function(array[i], array[f]) < 0)
			i++;
		j--;
		while(compare_function(array[j], array[f]) > 0)
			j--;
	}
	
	temp = array[f];
	array[f] = array[j];
	array[j] = temp;
	
	quickSort(array, f, j - 1, compare_function);
	quickSort(array, j + 1, l, compare_function);
}


/*
// some error checking code
*/

#if DEBUG_VALUE >= 1

void
dllistCheck(const DLList *that)
{
	int4 c = dllistLength(that);
	DLListNode *node;
	int4 i;
	
	for(node = that->start.next, i = 0; i < c && node; node = node->next, i++)
		;
	
	if(i != c)
		debugf("This list is too short in the forward direction.\n");
	
	if(node != &that->end)
		debugf("This list is messed up in the next links.\n");
	
	
	
	for(node = that->end.prev, i = 0; i < c && node; node = node->prev, i++)
		;
	
	if(i != c)
		debugf("This list is too short in the reverse direction.\n");
	
	if(node != &that->start)
		debugf("This list is messed up in the prev links.\n");
}

#endif
