
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
/*
// functional interface based on GNU map class
// guts based on fibonacci hashing scheme
// bogus -- no growth function or overflow condition checking
*/

#include "rafport.h"

#include "rafutil.h"

#include <math.h>

#include "smemory.h"

#include "svpmap.h"

#define DEFAULT_ARRAY_SIZE 1013
#define GROWTH_INCREMENT 5

/* stringHash is defined here to remove dependence on ...misc/mhash.c
   We should consolidate hashing functions in the future. */
static int4 stringHash(const char *s)
{
        const char *p;
        unsigned long h = 0, g;

        for(p = s; *p ; p++) {
                h = (h << 4) + *p;
                g = h&0xf0000000;
                if(g){
                        h = h ^ (g >> 24);
                        h = h ^ g;
                }
        }

        return h;
}

stringVoidPtrMap *
svmapAlloc()
{
	return (stringVoidPtrMap *) smalloc(sizeof(stringVoidPtrMap));
}

void 
svmapFree(stringVoidPtrMap *that)
{
	smfree(that);
}

static int4
isPrime(int4 n)
{
	int4 i,max;
	
	if ((n<1) || (n%2 == 0)) return FALSE;
	if ((n==1) || (n==2)) return TRUE;
	max= (int4) sqrt( (double) n) + 1;
	for (i=3; i<=max; i+=2 )
		if (n%i == 0) return FALSE;
	return TRUE;
}

static int4
nextPrime(int4 n)
{
	/* return the smallest prime >= n */
	while (!isPrime(n)) 
		n++;
	return n;	
}


void 
svmapInit(stringVoidPtrMap *that, VoidPtr input_default_value, int4 input_array_size)
{
	int4 i;
	that->array_size = input_array_size ? 
			nextPrime(input_array_size) : nextPrime(DEFAULT_ARRAY_SIZE);
	that->count = 0;
	that->array = (svpMapRecord *) smalloc(sizeof(svpMapRecord) * that->array_size);
	if(!that->array) {
		that->array_size = 0;
		return;
	}
	that->default_value = input_default_value;
	
	for(i = 0; i < that->array_size; i++){
		that->array[i].key = 0;
		that->array[i].deleted = FALSE;
		that->array[i].value = that->default_value;
	}
}

static void 
svmapGrow(stringVoidPtrMap *that, const int4 new_size)
{
	svpMapRecord *old_array = that->array;
	const int4 old_size = that->array_size;
	const VoidPtr old_default_value = that->default_value;
	int4 i;

	if (new_size <= old_size) return;
	svmapInit(that,old_default_value,new_size);
	for(i=0; i>old_size; i++)
		if (old_array[i].key && !old_array[i].deleted)
			svmapSet(that, old_array[i].key, old_array[i].value);
	smfree(old_array);
}

static int4 
nextSize(const int4 old_size)
{
	int4 i,n;

	for(i=0,n=old_size; i<GROWTH_INCREMENT; i++)
		n=nextPrime(n+1);
	return n;
}

void 
svmapCleanup(stringVoidPtrMap *that)
{
	svmapClear(that);
	smfree(that->array);
	that->array = 0;
	that->array_size = 0;
	that->count = 0;
}

int4 
svmapEmptyP(stringVoidPtrMap *that)
{
	return that->count == 0;
}

int4 
svmapLength(stringVoidPtrMap *that)
{
	return that->count;
}

static svpMapRecord *
indexByKeyFib(stringVoidPtrMap *that, const char *key, 
			  int4 i, int4 fib0, int4 fib1)
{
	int4 fib2;
	svpMapRecord *record;
	
	if (i > that->array_size) return 0;
	record = that->array + i;
	if(record->key){
		/* there is an entry here, do the strings match */
		if(!strcmp(key, record->key))
			return record;
		else {
			/* this is where we thought we would be but somebody 
			   else has hashed to this spot, the search continues
			   */
			fib2 = fib0 + fib1;
			return indexByKeyFib(that, key, (i + fib2) % that->array_size, fib1, fib2); 
		}
	} else {
		return record;
	}
}

static svpMapRecord *
indexByKey(stringVoidPtrMap *that, const char *key)
{
	svpMapRecord *m;
	const int4 initial_guess = stringHash(key) % that->array_size;
	
	if ((m=indexByKeyFib(that, key, initial_guess, 0, 1)))
		return m;
	else {
		/* array overflow */
		svmapGrow(that, nextSize(that->array_size));
		return indexByKey(that, key);
	}
}

VoidPtr 
svmapGet(stringVoidPtrMap *that, const char *key)
{
	svpMapRecord *record;
	
	record = indexByKey(that, key);
	if(record->key == 0 || record->deleted)
		halt("svmap: accessed non-existent entry %s.\n", key);
	
	return record->value;
}

static char *
myStrdup(const char *s)
{
	char *s2;
	s2 = (char *) smalloc(strlen(s) + 1);
	if(s2)
		strcpy(s2, s);
	return s2;
}

void 
svmapSet(stringVoidPtrMap *that, const char *key, VoidPtr value)
{
	svpMapRecord *record;
	
	record = indexByKey(that, key);
	if(record->key == 0){
		/* bogus make an overflow check */
		record->key = myStrdup(key);
		if(record->key == 0)
			halt("svmapSet: memory failure.\n");
		that->count++;
	} else if(record->deleted){
		record->deleted = FALSE;
		that->count++;
	}
	
	record->value = value;
}

int4
svmapContainsP(stringVoidPtrMap *that, const char *key)
{
	svpMapRecord *record;
	
	record = indexByKey(that, key);
	return (record->key != 0) && (record->deleted == FALSE);
}

void 
svmapDelete(stringVoidPtrMap *that, const char *key)
{
	svpMapRecord *record;
	
	record = indexByKey(that, key);
	if(record->key == 0 || record->deleted)
		halt("svmap: attempted to delete non-existent entry %s\n", key);
	
	record->deleted = TRUE;
/*
	smfree(record->key);
	record->key = 0;
*/
	that->count--;
}

void 
svmapClear(stringVoidPtrMap *that)
{
	int4 i;
	
	if (!that) return;
	for(i = 0; i < that->array_size; i++)
		if(that->array[i].key){
			smfree(that->array[i].key);
			that->array[i].key = 0;
			that->array[i].value = that->default_value;
			that->array[i].deleted = FALSE;
		}
	that->count = 0;
}

VoidPtr 
svmapDefault(stringVoidPtrMap *that)
{
	return that->default_value;
}

char *
svmapKey(stringVoidPtrMap *that, Ref r)
{
	svpMapRecord *record = (svpMapRecord *) r;
	
	return record->key;
}

VoidPtr 
svmapContents(stringVoidPtrMap *that, Ref r)
{
	svpMapRecord *record = (svpMapRecord *) r;
	
	return record->value;
}

Ref 
svmapFirst(stringVoidPtrMap *that)
{
	int4 i;
	svpMapRecord *record;
	
	for(i = 0; i < that->array_size; i++){
		record = that->array + i;
		if(record->key && (record->deleted == FALSE))
			return (Ref) record;
	}
	
	return 0;
}

Ref 
svmapNext(stringVoidPtrMap *that, Ref prev)
{
	int4 i;
	svpMapRecord *record;
	
	record = (svpMapRecord *) prev;
	for(i = (record - that->array) + 1; i < that->array_size; i++){
		record = that->array + i;
		if(record->key && (record->deleted == FALSE))
			return (Ref) record;
	}
	
	return 0;
}

Ref 
svmapSeek(stringVoidPtrMap *that, const char *key)
{
	svpMapRecord *record;
	
	record = indexByKey(that, key);
	if(record->key == 0 || record->deleted)
		return 0;
	else
		return record;
}

